from pathlib import Path

import pytest
from fastmcp import Client

from protein_quest.mcp_server import mcp


@pytest.fixture
def cif_path() -> Path:
    return Path(__file__).parent / "pdbe" / "fixtures" / "2y29.cif"


@pytest.mark.asyncio
async def test_nr_residues_in_chain(cif_path: Path):
    async with Client(mcp) as client:
        result = await client.call_tool("nr_residues_in_chain", {"file": cif_path})
        assert result.data == 8



"""Module for dealing with secondary structure."""

import logging
from collections.abc import Generator, Iterable
from dataclasses import dataclass
from pathlib import Path

from gemmi import Structure, read_structure, set_leak_warnings

from protein_quest.converter import PositiveInt, Ratio, converter

logger = logging.getLogger(__name__)

# TODO remove once v0.7.4 of gemmi is released,
# as uv pip install git+https://github.com/project-gemmi/gemmi.git installs 0.7.4.dev0 which does not print leaks
# Swallow gemmi leaked function warnings
set_leak_warnings(False)

# TODO if a structure has no secondary structure information, calculate it with `gemmi ss`.
# https://github.com/MonomerLibrary/monomers/wiki/Installation as --monomers dir
# gemmi executable is in https://pypi.org/project/gemmi-program/
# `gemmi ss` only prints secondary structure to stdout with `-v` flag.


def nr_of_residues_in_total(structure: Structure) -> int:
    """Count the total number of residues in the structure.

    Args:
        structure: The gemmi Structure object to analyze.

    Returns:
        The total number of residues in the structure.
    """
    count = 0
    for model in structure:
        for chain in model:
            count += len(chain)
    return count


def nr_of_residues_in_helix(structure: Structure) -> int:
    """Count the number of residues in alpha helices.

    Requires structure to have secondary structure information.

    Args:
        structure: The gemmi Structure object to analyze.

    Returns:
        The number of residues in alpha helices.
    """
    # For cif files from AlphaFold the helix.length is set to -1
    # so use resid instead
    count = 0
    for helix in structure.helices:
        end = helix.end.res_id.seqid.num
        start = helix.start.res_id.seqid.num
        if end is None or start is None:
            logger.warning(f"Invalid helix coordinates: {helix.end} or {helix.start}")
            continue
        length = end - start + 1
        count += length
    return count


def nr_of_residues_in_sheet(structure: Structure) -> int:
    """Count the number of residues in beta sheets.

    Requires structure to have secondary structure information.

    Args:
        structure: The gemmi Structure object to analyze.

    Returns:
        The number of residues in beta sheets.
    """
    count = 0
    for sheet in structure.sheets:
        for strand in sheet.strands:
            end = strand.end.res_id.seqid.num
            start = strand.start.res_id.seqid.num
            if end is None or start is None:
                logger.warning(f"Invalid strand coordinates: {strand.end} or {strand.start}")
                continue
            length = end - start + 1
            count += length
    return count


@dataclass
class SecondaryStructureFilterQuery:
    """Query object to filter on secondary structure.

    Parameters:
        abs_min_helix_residues: Minimum number of residues in helices (absolute).
        abs_max_helix_residues: Maximum number of residues in helices (absolute).
        abs_min_sheet_residues: Minimum number of residues in sheets (absolute).
        abs_max_sheet_residues: Maximum number of residues in sheets (absolute).
        ratio_min_helix_residues: Minimum number of residues in helices (relative).
        ratio_max_helix_residues: Maximum number of residues in helices (relative).
        ratio_min_sheet_residues: Minimum number of residues in sheets (relative).
        ratio_max_sheet_residues: Maximum number of residues in sheets (relative).
    """

    abs_min_helix_residues: PositiveInt | None = None
    abs_max_helix_residues: PositiveInt | None = None
    abs_min_sheet_residues: PositiveInt | None = None
    abs_max_sheet_residues: PositiveInt | None = None
    ratio_min_helix_residues: Ratio | None = None
    ratio_max_helix_residues: Ratio | None = None
    ratio_min_sheet_residues: Ratio | None = None
    ratio_max_sheet_residues: Ratio | None = None

    def is_actionable(self) -> bool:
        """Check if the secondary structure query has any actionable filters.

        Returns:
            True if any of the filters are set, False otherwise.
        """
        return any(
            field is not None
            for field in [
                self.abs_min_helix_residues,
                self.abs_max_helix_residues,
                self.abs_min_sheet_residues,
                self.abs_max_sheet_residues,
                self.ratio_min_helix_residues,
                self.ratio_max_helix_residues,
                self.ratio_min_sheet_residues,
                self.ratio_max_sheet_residues,
            ]
        )


def _check_range(min_val, max_val, label):
    if min_val is not None and max_val is not None and min_val >= max_val:
        msg = f"Invalid {label} range: min {min_val} must be smaller than max {max_val}"
        raise ValueError(msg)


base_query_hook = converter.get_structure_hook(SecondaryStructureFilterQuery)


@converter.register_structure_hook
def secondary_structure_filter_query_hook(value, _type) -> SecondaryStructureFilterQuery:
    result: SecondaryStructureFilterQuery = base_query_hook(value, _type)
    _check_range(result.abs_min_helix_residues, result.abs_max_helix_residues, "absolute helix residue")
    _check_range(result.abs_min_sheet_residues, result.abs_max_sheet_residues, "absolute sheet residue")
    _check_range(result.ratio_min_helix_residues, result.ratio_max_helix_residues, "ratio helix residue")
    _check_range(result.ratio_min_sheet_residues, result.ratio_max_sheet_residues, "ratio sheet residue")
    return result


@dataclass
class SecondaryStructureStats:
    """Statistics about the secondary structure of a protein.

    Parameters:
        nr_residues: Total number of residues in the structure.
        nr_helix_residues: Number of residues in helices.
        nr_sheet_residues: Number of residues in sheets.
        helix_ratio: Ratio of residues in helices.
        sheet_ratio: Ratio of residues in sheets.
    """

    nr_residues: PositiveInt
    nr_helix_residues: PositiveInt
    nr_sheet_residues: PositiveInt
    helix_ratio: Ratio
    sheet_ratio: Ratio


@dataclass
class SecondaryStructureFilterResult:
    """Result of filtering on secondary structure.

    Parameters:
        stats: The secondary structure statistics.
        passed: Whether the structure passed the filtering criteria.
    """

    stats: SecondaryStructureStats
    passed: bool = False


def _gather_stats(structure: Structure) -> SecondaryStructureStats:
    nr_total_residues = nr_of_residues_in_total(structure)
    nr_helix_residues = nr_of_residues_in_helix(structure)
    nr_sheet_residues = nr_of_residues_in_sheet(structure)
    if nr_total_residues == 0:
        msg = "Structure has zero residues; cannot compute secondary structure ratios."
        raise ValueError(msg)
    helix_ratio = nr_helix_residues / nr_total_residues
    sheet_ratio = nr_sheet_residues / nr_total_residues
    return SecondaryStructureStats(
        nr_residues=nr_total_residues,
        nr_helix_residues=nr_helix_residues,
        nr_sheet_residues=nr_sheet_residues,
        helix_ratio=helix_ratio,
        sheet_ratio=sheet_ratio,
    )


def filter_on_secondary_structure(
    structure: Structure,
    query: SecondaryStructureFilterQuery,
) -> SecondaryStructureFilterResult:
    """Filter a structure based on secondary structure criteria.

    Args:
        structure: The gemmi Structure object to analyze.
        query: The filtering criteria to apply.

    Returns:
        Filtering statistics and whether structure passed.
    """
    stats = _gather_stats(structure)
    conditions: list[bool] = []

    # Helix absolute thresholds
    if query.abs_min_helix_residues is not None:
        conditions.append(stats.nr_helix_residues >= query.abs_min_helix_residues)
    if query.abs_max_helix_residues is not None:
        conditions.append(stats.nr_helix_residues <= query.abs_max_helix_residues)

    # Helix ratio thresholds
    if query.ratio_min_helix_residues is not None:
        conditions.append(stats.helix_ratio >= query.ratio_min_helix_residues)
    if query.ratio_max_helix_residues is not None:
        conditions.append(stats.helix_ratio <= query.ratio_max_helix_residues)

    # Sheet absolute thresholds
    if query.abs_min_sheet_residues is not None:
        conditions.append(stats.nr_sheet_residues >= query.abs_min_sheet_residues)
    if query.abs_max_sheet_residues is not None:
        conditions.append(stats.nr_sheet_residues <= query.abs_max_sheet_residues)

    # Sheet ratio thresholds
    if query.ratio_min_sheet_residues is not None:
        conditions.append(stats.sheet_ratio >= query.ratio_min_sheet_residues)
    if query.ratio_max_sheet_residues is not None:
        conditions.append(stats.sheet_ratio <= query.ratio_max_sheet_residues)

    if not conditions:
        msg = "No filtering conditions provided. Please specify at least one condition."
        raise ValueError(msg)
    passed = all(conditions)
    return SecondaryStructureFilterResult(stats=stats, passed=passed)


def filter_file_on_secondary_structure(
    file_path: Path,
    query: SecondaryStructureFilterQuery,
) -> SecondaryStructureFilterResult:
    """Filter a structure file based on secondary structure criteria.

    Args:
        file_path: The path to the structure file to analyze.
        query: The filtering criteria to apply.

    Returns:
        Filtering statistics and whether file passed.
    """
    structure = read_structure(str(file_path))
    return filter_on_secondary_structure(structure, query)


def filter_files_on_secondary_structure(
    file_paths: Iterable[Path],
    query: SecondaryStructureFilterQuery,
) -> Generator[tuple[Path, SecondaryStructureFilterResult]]:
    """Filter multiple structure files based on secondary structure criteria.

    Args:
        file_paths: A list of paths to the structure files to analyze.
        query: The filtering criteria to apply.

    Yields:
        For each file returns the filtering statistics and whether structure passed.
    """
    # TODO check if quick enough in serial mode, if not switch to dask map
    for file_path in file_paths:
        result = filter_file_on_secondary_structure(file_path, query)
        yield file_path, result



"""Module for fetching structures from PDBe."""

from collections.abc import Iterable, Mapping
from pathlib import Path

from protein_quest.utils import Cacher, retrieve_files, run_async


def _map_id_mmcif(pdb_id: str) -> tuple[str, str]:
    """
    Map PDB id to a download gzipped mmCIF url and file.

    For example for PDB id "8WAS", the url will be
    "https://www.ebi.ac.uk/pdbe/entry-files/download/8was.cif.gz" and the file will be "8was.cif.gz".

    Args:
        pdb_id: The PDB ID to map.

    Returns:
        A tuple containing the URL to download the mmCIF file and the filename.
    """
    fn = f"{pdb_id.lower()}.cif.gz"
    # On PDBe you can sometimes download an updated mmCIF file,
    # Current url is for the archive mmCIF file
    # TODO check if archive is OK, or if we should try to download the updated file
    # this will cause many more requests, so we should only do this if needed
    url = f"https://www.ebi.ac.uk/pdbe/entry-files/download/{fn}"
    return url, fn


async def fetch(
    ids: Iterable[str], save_dir: Path, max_parallel_downloads: int = 5, cacher: Cacher | None = None
) -> Mapping[str, Path]:
    """Fetches mmCIF files from the PDBe database.

    Args:
        ids: A set of PDB IDs to fetch.
        save_dir: The directory to save the fetched mmCIF files to.
        max_parallel_downloads: The maximum number of parallel downloads.
        cacher: An optional cacher to use for caching downloaded files.

    Returns:
        A dict of id and paths to the downloaded mmCIF files.
    """

    # The future result, is in a different order than the input ids,
    # so we need to map the ids to the urls and filenames.

    id2urls = {pdb_id: _map_id_mmcif(pdb_id) for pdb_id in ids}
    urls = list(id2urls.values())
    id2paths = {pdb_id: save_dir / fn for pdb_id, (_, fn) in id2urls.items()}

    await retrieve_files(urls, save_dir, max_parallel_downloads, desc="Downloading PDBe mmCIF files", cacher=cacher)
    return id2paths


def sync_fetch(ids: Iterable[str], save_dir: Path, max_parallel_downloads: int = 5) -> Mapping[str, Path]:
    """Synchronously fetches mmCIF files from the PDBe database.

    Args:
        ids: A set of PDB IDs to fetch.
        save_dir: The directory to save the fetched mmCIF files to.
        max_parallel_downloads: The maximum number of parallel downloads.

    Returns:
        A dict of id and paths to the downloaded mmCIF files.
    """
    return run_async(fetch(ids, save_dir, max_parallel_downloads))



from protein_quest.alphafold.entry_summary import EntrySummary
from protein_quest.alphafold.fetch import converter


def test_loads_from_json_string():
    summary_json = '[{"toolUsed":"AlphaFold Monomer v2.0 pipeline","providerId":"GDM","entityType":"protein","isUniProt":true,"modelEntityId":"AF-A0A087WUV0-F1","modelCreatedDate":"2022-06-01T00:00:00Z","sequenceVersionDate":"2014-10-29T00:00:00Z","globalMetricValue":65.58,"fractionPlddtVeryLow":0.375,"fractionPlddtLow":0.082,"fractionPlddtConfident":0.284,"fractionPlddtVeryHigh":0.259,"latestVersion":4,"allVersions":[1,2,3,4],"sequence":"MEPEGRGSLFEDSDLLHAGNPKENDVTAVLLTPGSQELMIRDMAEALTQWRQLNSPQGDVPEKPRNLVLLGLPISTPDVISQLEHEEELEREVSKAASQKHWETIPESKELTPEKDISEEESAPGVLIVRFSKESSSECEDSLESQQENHEKHLIQEAVTEKSSRERSYQSDEFRRNCTQRSLLVQQQGERLHHCDSFKNNLKQNSDIIRHERICAGKKPWKCNECEKAFSYYSAFVLHQRIHTGEKPYECNECGKAFSQSIHLTLHQRIHTGEKPYECHECGKAFSHRSALIRHHIIHTGEKPYECNECGKAFNQSSYLTQHQRIHTGEKPYECNECGKAFSQSTFLTQHQVIHTGEKPYKCNECGKAFSDRSGLIQHQRTHTGERPYECNECGKAFGYCSALTQHQRTHTGEKPYKCNDCAKAFSDRSALIRHQRTHTGEKPYKCKDCGKAFSQSSSLTKHQKTHTGEKPYKCKECGKAFSQSSSLSQHQKTHAGVKTKKYVQALSEHLTFGQHKRIHTG","sequenceStart":1,"sequenceEnd":522,"sequenceChecksum":"5DE83E4BE25B68BD","isUniProtReviewed":false,"uniprotAccession":"A0A087WUV0","uniprotId":"A0A087WUV0_HUMAN","uniprotDescription":"Uncharacterized protein","taxId":9606,"organismScientificName":"Homo sapiens","isUniProtReferenceProteome":true,"bcifUrl":"https://alphafold.ebi.ac.uk/files/AF-A0A087WUV0-F1-model_v4.bcif","cifUrl":"https://alphafold.ebi.ac.uk/files/AF-A0A087WUV0-F1-model_v4.cif","pdbUrl":"https://alphafold.ebi.ac.uk/files/AF-A0A087WUV0-F1-model_v4.pdb","paeImageUrl":"https://alphafold.ebi.ac.uk/files/AF-A0A087WUV0-F1-predicted_aligned_error_v4.png","paeDocUrl":"https://alphafold.ebi.ac.uk/files/AF-A0A087WUV0-F1-predicted_aligned_error_v4.json","amAnnotationsUrl":"https://alphafold.ebi.ac.uk/files/AF-A0A087WUV0-F1-aa-substitutions.csv","amAnnotationsHg38Url":"https://alphafold.ebi.ac.uk/files/AF-A0A087WUV0-F1-hg38.csv","entryId":"AF-A0A087WUV0-F1","uniprotSequence":"MEPEGRGSLFEDSDLLHAGNPKENDVTAVLLTPGSQELMIRDMAEALTQWRQLNSPQGDVPEKPRNLVLLGLPISTPDVISQLEHEEELEREVSKAASQKHWETIPESKELTPEKDISEEESAPGVLIVRFSKESSSECEDSLESQQENHEKHLIQEAVTEKSSRERSYQSDEFRRNCTQRSLLVQQQGERLHHCDSFKNNLKQNSDIIRHERICAGKKPWKCNECEKAFSYYSAFVLHQRIHTGEKPYECNECGKAFSQSIHLTLHQRIHTGEKPYECHECGKAFSHRSALIRHHIIHTGEKPYECNECGKAFNQSSYLTQHQRIHTGEKPYECNECGKAFSQSTFLTQHQVIHTGEKPYKCNECGKAFSDRSGLIQHQRTHTGERPYECNECGKAFGYCSALTQHQRTHTGEKPYKCNDCAKAFSDRSALIRHQRTHTGEKPYKCKDCGKAFSQSSSLTKHQKTHTGEKPYKCKECGKAFSQSSSLSQHQKTHAGVKTKKYVQALSEHLTFGQHKRIHTG","uniprotStart":1,"uniprotEnd":522,"isReferenceProteome":true,"isReviewed":false}]'

    result = converter.loads(summary_json, list[EntrySummary])

    assert isinstance(result, list)
    assert all(isinstance(item, EntrySummary) for item in result)
    assert result[0].uniprotAccession == "A0A087WUV0"



from pathlib import Path

import pytest

from protein_quest.alphafold.fetch import fetch_many


@pytest.mark.vcr
def test_fetch_many(tmp_path: Path):
    theid = "P05067"
    ids = [theid]

    results = fetch_many(ids, tmp_path, {"summary", "pdb"})

    assert len(results) == 1
    fresult = results[0]
    assert fresult.uniprot_acc == theid
    assert fresult.summary is not None
    assert (tmp_path / f"{theid}.json").exists()
    assert fresult.pdb_file and fresult.pdb_file.exists()



from pathlib import Path

import pytest

from protein_quest.emdb import fetch


@pytest.mark.asyncio
@pytest.mark.vcr
async def test_fetch(tmp_path: Path):
    # use small emdb entry
    emdb_ids = ["EMD-1470"]

    results = await fetch(emdb_ids, tmp_path)
    expected = {"EMD-1470": tmp_path / "emd_1470.map.gz"}
    assert results == expected
    assert all(path.exists() for path in results.values())



from pathlib import Path

import gemmi
import pytest

from protein_quest.alphafold.confidence import (
    ConfidenceFilterQuery,
    ConfidenceFilterResult,
    converter,
    filter_files_on_confidence,
    filter_out_low_confidence_residues,
    find_high_confidence_residues,
)
from protein_quest.pdbe.io import nr_residues_in_chain


@pytest.fixture
def sample_pdb_file() -> Path:
    return Path(__file__).parent / "AF-A1YPR0-F1-model_v4.pdb"


@pytest.fixture
def sample_pdb(sample_pdb_file: Path) -> gemmi.Structure:
    return gemmi.read_structure(str(sample_pdb_file))


def test_find_high_confidence_residues(sample_pdb: gemmi.Structure):
    residues = list(find_high_confidence_residues(sample_pdb, 90))

    assert len(residues) == 22


def test_filter_out_low_confidence_residues(sample_pdb: gemmi.Structure):
    # Make sure we start with >22 residues
    assert len(sample_pdb[0][0]) == 619

    residues = set(find_high_confidence_residues(sample_pdb, 90))
    new_structure = filter_out_low_confidence_residues(sample_pdb, residues)

    assert len(new_structure[0][0]) == 22


def test_filter_files_on_confidence(sample_pdb_file: Path, tmp_path: Path):
    input_files = [sample_pdb_file]
    query = ConfidenceFilterQuery(
        confidence=90,
        max_residues=40,
        min_residues=10,
    )

    results = list(filter_files_on_confidence(input_files, query, tmp_path))

    expected = [
        ConfidenceFilterResult(
            input_file=sample_pdb_file.name,
            count=22,
            filtered_file=tmp_path / sample_pdb_file.name,
        )
    ]

    assert results == expected
    assert results[0].filtered_file is not None
    assert results[0].filtered_file.exists()
    assert nr_residues_in_chain(results[0].filtered_file) == 22


def test_query_converter():
    result = converter.structure(
        {
            "confidence": 90.0,
            "min_residues": 10,
            "max_residues": 100,
        },
        ConfidenceFilterQuery,
    )
    expected = ConfidenceFilterQuery(
        confidence=90.0,
        min_residues=10,
        max_residues=100,
    )
    assert result == expected


@pytest.mark.parametrize(
    "raw,match",
    [
        (
            {
                "confidence": 42,
                "min_residues": -10,
                "max_residues": 100,
            },
            "is not a valid positive integer",
        ),
        (
            {
                "confidence": 1234,
                "min_residues": 10,
                "max_residues": 100,
            },
            "is not a valid percentage",
        ),
        (
            {
                "confidence": -10,
                "min_residues": 10,
                "max_residues": 100,
            },
            "is not a valid percentage",
        ),
        (
            {
                "confidence": "not a number",
                "min_residues": 10,
                "max_residues": 100,
            },
            "could not convert string to float",
        ),
        (
            {
                "confidence": 42,
                "min_residues": "not a number",
                "max_residues": 100,
            },
            "invalid literal for int",
        ),
    ],
)
def test_query_converter_bad_confidence(raw: dict, match: str):
    with pytest.RaisesGroup(pytest.RaisesExc(ValueError, match=match)):
        converter.structure(
            raw,
            ConfidenceFilterQuery,
        )


@pytest.mark.parametrize(
    "raw,match",
    [
        (
            {
                "confidence": 1,
                "min_residues": 80,
                "max_residues": 20,
            },
            "cannot be larger than",
        ),
    ],
)
def test_query_converter_bad_residues(raw: dict, match: str):
    with pytest.raises(ValueError, match=match):
        converter.structure(
            raw,
            ConfidenceFilterQuery,
        )



from pathlib import Path

import pytest

from protein_quest.pdbe.fetch import fetch, sync_fetch


@pytest.mark.asyncio
@pytest.mark.vcr
async def test_fetch(tmp_path: Path):
    theid = "2Y29"
    ids = [theid]

    results = await fetch(ids, tmp_path)

    expected = {theid: tmp_path / f"{theid.lower()}.cif.gz"}
    assert results == expected


@pytest.mark.default_cassette("test_fetch.yaml")  # pyrefly: ignore[not-callable]
@pytest.mark.vcr
def test_sync_fetch(tmp_path: Path):
    theid = "2Y29"
    ids = [theid]

    results = sync_fetch(ids, tmp_path)

    expected = {theid: tmp_path / f"{theid.lower()}.cif.gz"}
    assert results == expected



"""Module for cli parsers and handlers."""

import argparse
import asyncio
import csv
import logging
import os
import sys
from collections.abc import Callable, Generator, Iterable
from importlib.util import find_spec
from io import TextIOWrapper
from pathlib import Path
from textwrap import dedent

from cattrs import structure
from rich import print as rprint
from rich.logging import RichHandler
from rich.markdown import Markdown
from rich.panel import Panel
from rich_argparse import ArgumentDefaultsRichHelpFormatter
from tqdm.rich import tqdm

from protein_quest.__version__ import __version__
from protein_quest.alphafold.confidence import ConfidenceFilterQuery, filter_files_on_confidence
from protein_quest.alphafold.fetch import DownloadableFormat, downloadable_formats
from protein_quest.alphafold.fetch import fetch_many as af_fetch
from protein_quest.converter import converter
from protein_quest.emdb import fetch as emdb_fetch
from protein_quest.filters import filter_files_on_chain, filter_files_on_residues
from protein_quest.go import Aspect, allowed_aspects, search_gene_ontology_term, write_go_terms_to_csv
from protein_quest.pdbe import fetch as pdbe_fetch
from protein_quest.pdbe.io import glob_structure_files, locate_structure_file
from protein_quest.ss import SecondaryStructureFilterQuery, filter_files_on_secondary_structure
from protein_quest.taxonomy import SearchField, _write_taxonomy_csv, search_fields, search_taxon
from protein_quest.uniprot import (
    ComplexPortalEntry,
    PdbResult,
    Query,
    search4af,
    search4emdb,
    search4interaction_partners,
    search4macromolecular_complexes,
    search4pdb,
    search4uniprot,
)
from protein_quest.utils import (
    Cacher,
    CopyMethod,
    DirectoryCacher,
    PassthroughCacher,
    copy_methods,
    copyfile,
    user_cache_root_dir,
)

logger = logging.getLogger(__name__)


def _add_search_uniprot_parser(subparsers: argparse._SubParsersAction):
    """Add search uniprot subcommand parser."""
    parser = subparsers.add_parser(
        "uniprot",
        help="Search UniProt accessions",
        description="Search for UniProt accessions based on various criteria in the Uniprot SPARQL endpoint.",
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    parser.add_argument(
        "output",
        type=argparse.FileType("w", encoding="UTF-8"),
        help="Output text file for UniProt accessions (one per line). Use `-` for stdout.",
    )
    parser.add_argument("--taxon-id", type=str, help="NCBI Taxon ID, e.g. 9606 for Homo Sapiens")
    parser.add_argument(
        "--reviewed",
        action=argparse.BooleanOptionalAction,
        help="Reviewed=swissprot, no-reviewed=trembl. Default is uniprot=swissprot+trembl.",
        default=None,
    )
    parser.add_argument(
        "--subcellular-location-uniprot",
        type=str,
        help="Subcellular location label as used by UniProt (e.g. nucleus)",
    )
    parser.add_argument(
        "--subcellular-location-go",
        dest="subcellular_location_go",
        action="append",
        help="GO term(s) for subcellular location (e.g. GO:0005634). Can be given multiple times.",
    )
    parser.add_argument(
        "--molecular-function-go",
        dest="molecular_function_go",
        action="append",
        help="GO term(s) for molecular function (e.g. GO:0003677). Can be given multiple times.",
    )
    parser.add_argument("--limit", type=int, default=10_000, help="Maximum number of uniprot accessions to return")
    parser.add_argument("--timeout", type=int, default=1_800, help="Maximum seconds to wait for query to complete")


def _add_search_pdbe_parser(subparsers: argparse._SubParsersAction):
    """Add search pdbe subcommand parser."""
    parser = subparsers.add_parser(
        "pdbe",
        help="Search PDBe structures of given UniProt accessions",
        description="Search for PDB structures of given UniProt accessions in the Uniprot SPARQL endpoint.",
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    parser.add_argument(
        "uniprot_accs",
        type=argparse.FileType("r", encoding="UTF-8"),
        help="Text file with UniProt accessions (one per line). Use `-` for stdin.",
    )
    parser.add_argument(
        "output_csv",
        type=argparse.FileType("w", encoding="UTF-8"),
        help=dedent("""\
            Output CSV with `uniprot_acc`, `pdb_id`, `method`, `resolution`, `uniprot_chains`, `chain` columns.
            Where `uniprot_chains` is the raw UniProt chain string, for example `A=1-100`.
            and where `chain` is the first chain from `uniprot_chains`, for example `A`.
            Use `-` for stdout.
        """),
    )
    parser.add_argument(
        "--limit", type=int, default=10_000, help="Maximum number of PDB uniprot accessions combinations to return"
    )
    parser.add_argument("--timeout", type=int, default=1_800, help="Maximum seconds to wait for query to complete")


def _add_search_alphafold_parser(subparsers: argparse._SubParsersAction):
    """Add search alphafold subcommand parser."""
    parser = subparsers.add_parser(
        "alphafold",
        help="Search AlphaFold structures of given UniProt accessions",
        description="Search for AlphaFold structures of given UniProt accessions in the Uniprot SPARQL endpoint.",
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    parser.add_argument(
        "uniprot_accs",
        type=argparse.FileType("r", encoding="UTF-8"),
        help="Text file with UniProt accessions (one per line). Use `-` for stdin.",
    )
    parser.add_argument(
        "output_csv",
        type=argparse.FileType("w", encoding="UTF-8"),
        help="Output CSV with AlphaFold IDs per UniProt accession. Use `-` for stdout.",
    )
    parser.add_argument(
        "--limit", type=int, default=10_000, help="Maximum number of Alphafold entry identifiers to return"
    )
    parser.add_argument("--timeout", type=int, default=1_800, help="Maximum seconds to wait for query to complete")


def _add_search_emdb_parser(subparsers: argparse._SubParsersAction):
    """Add search emdb subcommand parser."""
    parser = subparsers.add_parser(
        "emdb",
        help="Search Electron Microscopy Data Bank (EMDB) identifiers of given UniProt accessions",
        description=dedent("""\
            Search for Electron Microscopy Data Bank (EMDB) identifiers of given UniProt accessions
            in the Uniprot SPARQL endpoint.
        """),
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    parser.add_argument(
        "uniprot_accs",
        type=argparse.FileType("r", encoding="UTF-8"),
        help="Text file with UniProt accessions (one per line). Use `-` for stdin.",
    )
    parser.add_argument(
        "output_csv",
        type=argparse.FileType("w", encoding="UTF-8"),
        help="Output CSV with EMDB IDs per UniProt accession. Use `-` for stdout.",
    )
    parser.add_argument("--limit", type=int, default=10_000, help="Maximum number of EMDB entry identifiers to return")
    parser.add_argument("--timeout", type=int, default=1_800, help="Maximum seconds to wait for query to complete")


def _add_search_go_parser(subparsers: argparse._SubParsersAction):
    """Add search go subcommand parser"""
    parser = subparsers.add_parser(
        "go",
        help="Search for Gene Ontology (GO) terms",
        description="Search for Gene Ontology (GO) terms in the EBI QuickGO API.",
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    parser.add_argument(
        "term",
        type=str,
        help="GO term to search for. For example `apoptosome`.",
    )
    parser.add_argument("--aspect", type=str, choices=allowed_aspects, help="Filter on aspect.")
    parser.add_argument(
        "output_csv",
        type=argparse.FileType("w", encoding="UTF-8"),
        help="Output CSV with GO term results. Use `-` for stdout.",
    )
    parser.add_argument("--limit", type=int, default=100, help="Maximum number of GO term results to return")


def _add_search_taxonomy_parser(subparser: argparse._SubParsersAction):
    """Add search taxonomy subcommand parser."""
    parser = subparser.add_parser(
        "taxonomy",
        help="Search for taxon information in UniProt",
        description=dedent("""\
            Search for taxon information in UniProt.
            Uses https://www.uniprot.org/taxonomy?query=*.
        """),
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    parser.add_argument(
        "query", type=str, help="Search query for the taxon. Surround multiple words with quotes (' or \")."
    )
    parser.add_argument(
        "output_csv",
        type=argparse.FileType("w", encoding="UTF-8"),
        help="Output CSV with taxonomy results. Use `-` for stdout.",
    )
    parser.add_argument(
        "--field",
        type=str,
        choices=search_fields,
        help=dedent("""\
            Field to search in. If not given then searches all fields.
            If "tax_id" then searches by taxon ID.
            If "parent" then given a parent taxon ID returns all its children.
            For example, if the parent taxon ID is 9606 (Human), it will return Neanderthal and Denisovan.
        """),
    )
    parser.add_argument("--limit", type=int, default=100, help="Maximum number of results to return")


def _add_search_interaction_partners_parser(subparsers: argparse._SubParsersAction):
    """Add search interaction partners subcommand parser."""
    parser = subparsers.add_parser(
        "interaction-partners",
        help="Search for interaction partners of given UniProt accession",
        description=dedent("""\
            Search for interaction partners of given UniProt accession
            in the Uniprot SPARQL endpoint and Complex Portal.
        """),
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    parser.add_argument(
        "uniprot_acc",
        type=str,
        help="UniProt accession (for example P12345).",
    )
    parser.add_argument(
        "--exclude",
        type=str,
        action="append",
        help="UniProt accessions to exclude from the results. For example already known interaction partners.",
    )
    parser.add_argument(
        "output_csv",
        type=argparse.FileType("w", encoding="UTF-8"),
        help="Output CSV with interaction partners per UniProt accession. Use `-` for stdout.",
    )
    parser.add_argument(
        "--limit", type=int, default=10_000, help="Maximum number of interaction partner uniprot accessions to return"
    )
    parser.add_argument("--timeout", type=int, default=1_800, help="Maximum seconds to wait for query to complete")


def _add_search_complexes_parser(subparsers: argparse._SubParsersAction):
    """Add search complexes subcommand parser."""
    description = dedent("""\
        Search for complexes in the Complex Portal.
        https://www.ebi.ac.uk/complexportal/

        The output CSV file has the following columns:

        - query_protein: UniProt accession used as query
        - complex_id: Complex Portal identifier
        - complex_url: URL to the Complex Portal entry
        - complex_title: Title of the complex
        - members: Semicolon-separated list of UniProt accessions of complex members
    """)
    parser = subparsers.add_parser(
        "complexes",
        help="Search for complexes in the Complex Portal",
        description=Markdown(description, style="argparse.text"),  # type: ignore using rich formatter makes this OK
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    parser.add_argument(
        "uniprot_accs",
        type=argparse.FileType("r", encoding="UTF-8"),
        help="Text file with UniProt accessions (one per line) as query for searching complexes. Use `-` for stdin.",
    )
    parser.add_argument(
        "output_csv",
        type=argparse.FileType("w", encoding="UTF-8"),
        help="Output CSV file with complex results. Use `-` for stdout.",
    )
    parser.add_argument("--limit", type=int, default=100, help="Maximum number of complex results to return")
    parser.add_argument("--timeout", type=int, default=1_800, help="Maximum seconds to wait for query to complete")


def _add_retrieve_pdbe_parser(subparsers: argparse._SubParsersAction):
    """Add retrieve pdbe subcommand parser."""
    parser = subparsers.add_parser(
        "pdbe",
        help="Retrieve PDBe gzipped mmCIF files for PDB IDs in CSV.",
        description=dedent("""\
            Retrieve mmCIF files from Protein Data Bank in Europe Knowledge Base (PDBe) website
            for unique PDB IDs listed in a CSV file.
        """),
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    parser.add_argument(
        "pdbe_csv",
        type=argparse.FileType("r", encoding="UTF-8"),
        help="CSV file with `pdb_id` column. Other columns are ignored. Use `-` for stdin.",
    )
    parser.add_argument("output_dir", type=Path, help="Directory to store downloaded PDBe mmCIF files")
    parser.add_argument(
        "--max-parallel-downloads",
        type=int,
        default=5,
        help="Maximum number of parallel downloads",
    )
    _add_cacher_arguments(parser)


def _add_retrieve_alphafold_parser(subparsers: argparse._SubParsersAction):
    """Add retrieve alphafold subcommand parser."""
    parser = subparsers.add_parser(
        "alphafold",
        help="Retrieve AlphaFold files for IDs in CSV",
        description="Retrieve AlphaFold files from the AlphaFold Protein Structure Database.",
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    parser.add_argument(
        "alphafold_csv",
        type=argparse.FileType("r", encoding="UTF-8"),
        help="CSV file with `af_id` column. Other columns are ignored. Use `-` for stdin.",
    )
    parser.add_argument("output_dir", type=Path, help="Directory to store downloaded AlphaFold files")
    parser.add_argument(
        "--what-formats",
        type=str,
        action="append",
        choices=sorted(downloadable_formats),
        help=dedent("""AlphaFold formats to retrieve. Can be specified multiple times.
            Default is 'summary' and 'cif'."""),
    )
    parser.add_argument(
        "--max-parallel-downloads",
        type=int,
        default=5,
        help="Maximum number of parallel downloads",
    )
    _add_cacher_arguments(parser)


def _add_retrieve_emdb_parser(subparsers: argparse._SubParsersAction):
    """Add retrieve emdb subcommand parser."""
    parser = subparsers.add_parser(
        "emdb",
        help="Retrieve Electron Microscopy Data Bank (EMDB) gzipped 3D volume files for EMDB IDs in CSV.",
        description=dedent("""\
            Retrieve volume files from Electron Microscopy Data Bank (EMDB) website
            for unique EMDB IDs listed in a CSV file.
        """),
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    parser.add_argument(
        "emdb_csv",
        type=argparse.FileType("r", encoding="UTF-8"),
        help="CSV file with `emdb_id` column. Other columns are ignored. Use `-` for stdin.",
    )
    parser.add_argument("output_dir", type=Path, help="Directory to store downloaded EMDB volume files")
    _add_cacher_arguments(parser)


def _add_filter_confidence_parser(subparsers: argparse._SubParsersAction):
    """Add filter confidence subcommand parser."""
    parser = subparsers.add_parser(
        "confidence",
        help="Filter AlphaFold mmcif/PDB files by confidence",
        description=dedent("""\
            Filter AlphaFold mmcif/PDB files by confidence (plDDT).
            Passed files are written with residues below threshold removed."""),
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    parser.add_argument("input_dir", type=Path, help="Directory with AlphaFold mmcif/PDB files")
    parser.add_argument("output_dir", type=Path, help="Directory to write filtered mmcif/PDB files")
    parser.add_argument("--confidence-threshold", type=float, default=70, help="pLDDT confidence threshold (0-100)")
    parser.add_argument(
        "--min-residues", type=int, default=0, help="Minimum number of high-confidence residues a structure should have"
    )
    parser.add_argument(
        "--max-residues",
        type=int,
        default=10_000_000,
        help="Maximum number of high-confidence residues a structure should have",
    )
    parser.add_argument(
        "--write-stats",
        type=argparse.FileType("w", encoding="UTF-8"),
        help=dedent("""\
            Write filter statistics to file.
            In CSV format with `<input_file>,<residue_count>,<passed>,<output_file>` columns.
            Use `-` for stdout."""),
    )
    _add_copy_method_arguments(parser)


def _add_filter_chain_parser(subparsers: argparse._SubParsersAction):
    """Add filter chain subcommand parser."""
    parser = subparsers.add_parser(
        "chain",
        help="Filter on chain.",
        description=dedent("""\
            For each input PDB/mmCIF and chain combination
            write a PDB/mmCIF file with just the given chain
            and rename it to chain `A`.
            Filtering is done in parallel using a Dask cluster."""),
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    parser.add_argument(
        "chains",
        type=argparse.FileType("r", encoding="UTF-8"),
        help="CSV file with `pdb_id` and `chain` columns. Other columns are ignored.",
    )
    parser.add_argument(
        "input_dir",
        type=Path,
        help=dedent("""\
        Directory with PDB/mmCIF files.
        Expected filenames are `{pdb_id}.cif.gz`, `{pdb_id}.cif`, `{pdb_id}.pdb.gz` or `{pdb_id}.pdb`.
    """),
    )
    parser.add_argument(
        "output_dir",
        type=Path,
        help=dedent("""\
        Directory to write the single-chain PDB/mmCIF files. Output files are in same format as input files."""),
    )
    parser.add_argument(
        "--scheduler-address",
        help=dedent("""Address of the Dask scheduler to connect to.
            If not provided, will create a local cluster.
            If set to `sequential` will run tasks sequentially."""),
    )
    _add_copy_method_arguments(parser)


def _add_filter_residue_parser(subparsers: argparse._SubParsersAction):
    """Add filter residue subcommand parser."""
    parser = subparsers.add_parser(
        "residue",
        help="Filter PDB/mmCIF files by number of residues in chain A",
        description=dedent("""\
            Filter PDB/mmCIF files by number of residues in chain A.
        """),
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    parser.add_argument("input_dir", type=Path, help="Directory with PDB/mmCIF files (e.g., from 'filter chain')")
    parser.add_argument(
        "output_dir",
        type=Path,
        help=dedent("""\
        Directory to write filtered PDB/mmCIF files. Files are copied without modification.
    """),
    )
    parser.add_argument("--min-residues", type=int, default=0, help="Min residues in chain A")
    parser.add_argument("--max-residues", type=int, default=10_000_000, help="Max residues in chain A")
    parser.add_argument(
        "--write-stats",
        type=argparse.FileType("w", encoding="UTF-8"),
        help=dedent("""\
            Write filter statistics to file.
            In CSV format with `<input_file>,<residue_count>,<passed>,<output_file>` columns.
            Use `-` for stdout."""),
    )
    _add_copy_method_arguments(parser)


def _add_filter_ss_parser(subparsers: argparse._SubParsersAction):
    """Add filter secondary structure subcommand parser."""
    parser = subparsers.add_parser(
        "secondary-structure",
        help="Filter PDB/mmCIF files by secondary structure",
        description="Filter PDB/mmCIF files by secondary structure",
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    parser.add_argument("input_dir", type=Path, help="Directory with PDB/mmCIF files (e.g., from 'filter chain')")
    parser.add_argument(
        "output_dir",
        type=Path,
        help=dedent("""\
            Directory to write filtered PDB/mmCIF files. Files are copied without modification.
        """),
    )
    parser.add_argument("--abs-min-helix-residues", type=int, help="Min residues in helices")
    parser.add_argument("--abs-max-helix-residues", type=int, help="Max residues in helices")
    parser.add_argument("--abs-min-sheet-residues", type=int, help="Min residues in sheets")
    parser.add_argument("--abs-max-sheet-residues", type=int, help="Max residues in sheets")
    parser.add_argument("--ratio-min-helix-residues", type=float, help="Min residues in helices (relative)")
    parser.add_argument("--ratio-max-helix-residues", type=float, help="Max residues in helices (relative)")
    parser.add_argument("--ratio-min-sheet-residues", type=float, help="Min residues in sheets (relative)")
    parser.add_argument("--ratio-max-sheet-residues", type=float, help="Max residues in sheets (relative)")
    parser.add_argument(
        "--write-stats",
        type=argparse.FileType("w", encoding="UTF-8"),
        help=dedent("""
            Write filter statistics to file. In CSV format with columns:
            `<input_file>,<nr_residues>,<nr_helix_residues>,<nr_sheet_residues>,
            <helix_ratio>,<sheet_ratio>,<passed>,<output_file>`.
            Use `-` for stdout.
        """),
    )
    _add_copy_method_arguments(parser)


def _add_search_subcommands(subparsers: argparse._SubParsersAction):
    """Add search command and its subcommands."""
    parser = subparsers.add_parser(
        "search",
        help="Search data sources",
        description="Search various things online.",
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    subsubparsers = parser.add_subparsers(dest="search_cmd", required=True)

    _add_search_uniprot_parser(subsubparsers)
    _add_search_pdbe_parser(subsubparsers)
    _add_search_alphafold_parser(subsubparsers)
    _add_search_emdb_parser(subsubparsers)
    _add_search_go_parser(subsubparsers)
    _add_search_taxonomy_parser(subsubparsers)
    _add_search_interaction_partners_parser(subsubparsers)
    _add_search_complexes_parser(subsubparsers)


def _add_retrieve_subcommands(subparsers: argparse._SubParsersAction):
    """Add retrieve command and its subcommands."""
    parser = subparsers.add_parser(
        "retrieve",
        help="Retrieve structure files",
        description="Retrieve structure files from online resources.",
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    subsubparsers = parser.add_subparsers(dest="retrieve_cmd", required=True)

    _add_retrieve_pdbe_parser(subsubparsers)
    _add_retrieve_alphafold_parser(subsubparsers)
    _add_retrieve_emdb_parser(subsubparsers)


def _add_filter_subcommands(subparsers: argparse._SubParsersAction):
    """Add filter command and its subcommands."""
    parser = subparsers.add_parser("filter", help="Filter files", formatter_class=ArgumentDefaultsRichHelpFormatter)
    subsubparsers = parser.add_subparsers(dest="filter_cmd", required=True)

    _add_filter_confidence_parser(subsubparsers)
    _add_filter_chain_parser(subsubparsers)
    _add_filter_residue_parser(subsubparsers)
    _add_filter_ss_parser(subsubparsers)


def _add_mcp_command(subparsers: argparse._SubParsersAction):
    """Add MCP command."""

    parser = subparsers.add_parser(
        "mcp",
        help="Run Model Context Protocol (MCP) server",
        description=(
            "Run Model Context Protocol (MCP) server. "
            "Can be used by agentic LLMs like Claude Sonnet 4 as a set of tools."
        ),
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    parser.add_argument(
        "--transport", default="stdio", choices=["stdio", "http", "streamable-http"], help="Transport protocol to use"
    )
    parser.add_argument("--host", default="127.0.0.1", help="Host to bind the server to")
    parser.add_argument("--port", default=8000, type=int, help="Port to bind the server to")


def _add_copy_method_arguments(parser):
    parser.add_argument(
        "--copy-method",
        type=str,
        choices=copy_methods,
        default="hardlink",
        help=dedent("""\
            How to make target file be same file as source file.
            By default uses hardlinks to save disk space.
            Note that hardlinks only work within the same filesystem and are harder to track.
            If you want to track cached files easily then use 'symlink'.
            On Windows you need developer mode or admin privileges to create symlinks.
        """),
    )


def _add_cacher_arguments(parser: argparse.ArgumentParser):
    """Add cacher arguments to parser."""
    parser.add_argument(
        "--no-cache",
        action="store_true",
        help="Disable caching of files to central location.",
    )
    parser.add_argument(
        "--cache-dir",
        type=Path,
        default=user_cache_root_dir(),
        help="Directory to use as cache for files.",
    )
    _add_copy_method_arguments(parser)


def make_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Protein Quest CLI", prog="protein-quest", formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    parser.add_argument("--log-level", default="WARNING", choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"])
    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")

    subparsers = parser.add_subparsers(dest="command", required=True)

    _add_search_subcommands(subparsers)
    _add_retrieve_subcommands(subparsers)
    _add_filter_subcommands(subparsers)
    _add_mcp_command(subparsers)

    return parser


def main():
    """Main entry point for the CLI."""
    parser = make_parser()
    args = parser.parse_args()
    logging.basicConfig(level=args.log_level, handlers=[RichHandler(show_level=False)])

    # Dispatch table to reduce complexity
    cmd = args.command
    sub = getattr(args, f"{cmd}_cmd", None)
    handler = HANDLERS.get((cmd, sub))
    if handler is None:
        msg = f"Unknown command: {cmd} {sub}"
        raise SystemExit(msg)
    handler(args)


def _handle_search_uniprot(args):
    taxon_id = args.taxon_id
    reviewed = args.reviewed
    subcellular_location_uniprot = args.subcellular_location_uniprot
    subcellular_location_go = args.subcellular_location_go
    molecular_function_go = args.molecular_function_go
    limit = args.limit
    timeout = args.timeout
    output_file = args.output

    query = structure(
        {
            "taxon_id": taxon_id,
            "reviewed": reviewed,
            "subcellular_location_uniprot": subcellular_location_uniprot,
            "subcellular_location_go": subcellular_location_go,
            "molecular_function_go": molecular_function_go,
        },
        Query,
    )
    rprint("Searching for UniProt accessions")
    accs = search4uniprot(query=query, limit=limit, timeout=timeout)
    rprint(f"Found {len(accs)} UniProt accessions, written to {output_file.name}")
    _write_lines(output_file, sorted(accs))


def _handle_search_pdbe(args):
    uniprot_accs = args.uniprot_accs
    limit = args.limit
    timeout = args.timeout
    output_csv = args.output_csv

    accs = set(_read_lines(uniprot_accs))
    rprint(f"Finding PDB entries for {len(accs)} uniprot accessions")
    results = search4pdb(accs, limit=limit, timeout=timeout)
    total_pdbs = sum([len(v) for v in results.values()])
    rprint(f"Found {total_pdbs} PDB entries for {len(results)} uniprot accessions")
    rprint(f"Written to {output_csv.name}")
    _write_pdbe_csv(output_csv, results)


def _handle_search_alphafold(args):
    uniprot_accs = args.uniprot_accs
    limit = args.limit
    timeout = args.timeout
    output_csv = args.output_csv

    accs = _read_lines(uniprot_accs)
    rprint(f"Finding AlphaFold entries for {len(accs)} uniprot accessions")
    results = search4af(accs, limit=limit, timeout=timeout)
    rprint(f"Found {len(results)} AlphaFold entries, written to {output_csv.name}")
    _write_dict_of_sets2csv(output_csv, results, "af_id")


def _handle_search_emdb(args):
    uniprot_accs = args.uniprot_accs
    limit = args.limit
    timeout = args.timeout
    output_csv = args.output_csv

    accs = _read_lines(uniprot_accs)
    rprint(f"Finding EMDB entries for {len(accs)} uniprot accessions")
    results = search4emdb(accs, limit=limit, timeout=timeout)
    total_emdbs = sum([len(v) for v in results.values()])
    rprint(f"Found {total_emdbs} EMDB entries, written to {output_csv.name}")
    _write_dict_of_sets2csv(output_csv, results, "emdb_id")


def _handle_search_go(args):
    term = structure(args.term, str)
    aspect: Aspect | None = args.aspect
    limit = structure(args.limit, int)
    output_csv: TextIOWrapper = args.output_csv

    if aspect:
        rprint(f"Searching for GO terms matching '{term}' with aspect '{aspect}'")
    else:
        rprint(f"Searching for GO terms matching '{term}'")
    results = asyncio.run(search_gene_ontology_term(term, aspect=aspect, limit=limit))
    rprint(f"Found {len(results)} GO terms, written to {output_csv.name}")
    write_go_terms_to_csv(results, output_csv)


def _handle_search_taxonomy(args):
    query: str = args.query
    field: SearchField | None = args.field
    limit: int = args.limit
    output_csv: TextIOWrapper = args.output_csv

    if field:
        rprint(f"Searching for taxon information matching '{query}' in field '{field}'")
    else:
        rprint(f"Searching for taxon information matching '{query}'")
    results = asyncio.run(search_taxon(query=query, field=field, limit=limit))
    rprint(f"Found {len(results)} taxons, written to {output_csv.name}")
    _write_taxonomy_csv(results, output_csv)


def _handle_search_interaction_partners(args: argparse.Namespace):
    uniprot_acc: str = args.uniprot_acc
    excludes: set[str] = set(args.exclude) if args.exclude else set()
    limit: int = args.limit
    timeout: int = args.timeout
    output_csv: TextIOWrapper = args.output_csv

    rprint(f"Searching for interaction partners of '{uniprot_acc}'")
    results = search4interaction_partners(uniprot_acc, excludes=excludes, limit=limit, timeout=timeout)
    rprint(f"Found {len(results)} interaction partners, written to {output_csv.name}")
    _write_lines(output_csv, results.keys())


def _handle_search_complexes(args: argparse.Namespace):
    uniprot_accs = args.uniprot_accs
    limit = args.limit
    timeout = args.timeout
    output_csv = args.output_csv

    accs = _read_lines(uniprot_accs)
    rprint(f"Finding complexes for {len(accs)} uniprot accessions")
    results = search4macromolecular_complexes(accs, limit=limit, timeout=timeout)
    rprint(f"Found {len(results)} complexes, written to {output_csv.name}")
    _write_complexes_csv(results, output_csv)


def _initialize_cacher(args: argparse.Namespace) -> Cacher:
    if args.no_cache:
        return PassthroughCacher()
    return DirectoryCacher(
        cache_dir=args.cache_dir,
        copy_method=args.copy_method,
    )


def _handle_retrieve_pdbe(args: argparse.Namespace):
    pdbe_csv = args.pdbe_csv
    output_dir = args.output_dir
    max_parallel_downloads = args.max_parallel_downloads
    cacher = _initialize_cacher(args)

    pdb_ids = _read_column_from_csv(pdbe_csv, "pdb_id")
    rprint(f"Retrieving {len(pdb_ids)} PDBe entries")
    result = asyncio.run(
        pdbe_fetch.fetch(pdb_ids, output_dir, max_parallel_downloads=max_parallel_downloads, cacher=cacher)
    )
    rprint(f"Retrieved {len(result)} PDBe entries")


def _handle_retrieve_alphafold(args):
    download_dir = args.output_dir
    what_formats = args.what_formats
    alphafold_csv = args.alphafold_csv
    max_parallel_downloads = args.max_parallel_downloads
    cacher = _initialize_cacher(args)

    if what_formats is None:
        what_formats = {"summary", "cif"}

    # TODO besides `uniprot_acc,af_id\n` csv also allow headless single column format
    #
    af_ids = _read_column_from_csv(alphafold_csv, "af_id")
    validated_what: set[DownloadableFormat] = structure(what_formats, set[DownloadableFormat])
    rprint(f"Retrieving {len(af_ids)} AlphaFold entries with formats {validated_what}")
    afs = af_fetch(
        af_ids, download_dir, what=validated_what, max_parallel_downloads=max_parallel_downloads, cacher=cacher
    )
    total_nr_files = sum(af.nr_of_files() for af in afs)
    rprint(f"Retrieved {total_nr_files} AlphaFold files and {len(afs)} summaries, written to {download_dir}")


def _handle_retrieve_emdb(args):
    emdb_csv = args.emdb_csv
    output_dir = args.output_dir
    cacher = _initialize_cacher(args)

    emdb_ids = _read_column_from_csv(emdb_csv, "emdb_id")
    rprint(f"Retrieving {len(emdb_ids)} EMDB entries")
    result = asyncio.run(emdb_fetch(emdb_ids, output_dir, cacher=cacher))
    rprint(f"Retrieved {len(result)} EMDB entries")


def _handle_filter_confidence(args: argparse.Namespace):
    # we are repeating types here and in add_argument call
    # TODO replace argparse with modern alternative like cyclopts
    # to get rid of duplication
    input_dir = structure(args.input_dir, Path)
    output_dir = structure(args.output_dir, Path)

    confidence_threshold = args.confidence_threshold
    min_residues = args.min_residues
    max_residues = args.max_residues
    stats_file: TextIOWrapper | None = args.write_stats
    copy_method: CopyMethod = structure(args.copy_method, CopyMethod)  # pyright: ignore[reportArgumentType]

    output_dir.mkdir(parents=True, exist_ok=True)
    input_files = sorted(glob_structure_files(input_dir))
    nr_input_files = len(input_files)
    rprint(f"Starting confidence filtering of {nr_input_files} mmcif/PDB files in {input_dir} directory.")
    query = converter.structure(
        {
            "confidence": confidence_threshold,
            "min_residues": min_residues,
            "max_residues": max_residues,
        },
        ConfidenceFilterQuery,
    )
    if stats_file:
        writer = csv.writer(stats_file)
        writer.writerow(["input_file", "residue_count", "passed", "output_file"])

    passed_count = 0
    for r in tqdm(
        filter_files_on_confidence(input_files, query, output_dir, copy_method=copy_method),
        total=len(input_files),
        unit="file",
    ):
        if r.filtered_file:
            passed_count += 1
        if stats_file:
            writer.writerow([r.input_file, r.count, r.filtered_file is not None, r.filtered_file])

    rprint(f"Filtered {passed_count} mmcif/PDB files by confidence, written to {output_dir} directory")
    if stats_file:
        rprint(f"Statistics written to {stats_file.name}")


def _handle_filter_chain(args):
    input_dir = args.input_dir
    output_dir = structure(args.output_dir, Path)
    pdb_id2chain_mapping_file = args.chains
    scheduler_address = structure(args.scheduler_address, str | None)  # pyright: ignore[reportArgumentType]
    copy_method: CopyMethod = structure(args.copy_method, CopyMethod)  # pyright: ignore[reportArgumentType]

    # make sure files in input dir with entries in mapping file are the same
    # complain when files from mapping file are missing on disk
    rows = list(_iter_csv_rows(pdb_id2chain_mapping_file))
    file2chain: set[tuple[Path, str]] = set()
    errors: list[FileNotFoundError] = []

    for row in rows:
        pdb_id = row["pdb_id"]
        chain = row["chain"]
        try:
            f = locate_structure_file(input_dir, pdb_id)
            file2chain.add((f, chain))
        except FileNotFoundError as e:
            errors.append(e)

    if errors:
        msg = f"Some structure files could not be found ({len(errors)} missing), skipping them"
        rprint(Panel(os.linesep.join(map(str, errors)), title=msg, style="red"))

    if not file2chain:
        rprint("[red]No valid structure files found. Exiting.")
        sys.exit(1)

    results = filter_files_on_chain(
        file2chain, output_dir, scheduler_address=scheduler_address, copy_method=copy_method
    )

    nr_written = len([r for r in results if r.passed])

    rprint(f"Wrote {nr_written} single-chain PDB/mmCIF files to {output_dir}.")

    for result in results:
        if result.discard_reason:
            rprint(f"[red]Discarding {result.input_file} ({result.discard_reason})[/red]")


def _handle_filter_residue(args):
    input_dir = structure(args.input_dir, Path)
    output_dir = structure(args.output_dir, Path)
    min_residues = structure(args.min_residues, int)
    max_residues = structure(args.max_residues, int)
    copy_method: CopyMethod = structure(args.copy_method, CopyMethod)  # pyright: ignore[reportArgumentType]
    stats_file: TextIOWrapper | None = args.write_stats

    if stats_file:
        writer = csv.writer(stats_file)
        writer.writerow(["input_file", "residue_count", "passed", "output_file"])

    nr_passed = 0
    input_files = sorted(glob_structure_files(input_dir))
    nr_total = len(input_files)
    rprint(f"Filtering {nr_total} files in {input_dir} directory by number of residues in chain A.")
    for r in filter_files_on_residues(
        input_files, output_dir, min_residues=min_residues, max_residues=max_residues, copy_method=copy_method
    ):
        if stats_file:
            writer.writerow([r.input_file, r.residue_count, r.passed, r.output_file])
        if r.passed:
            nr_passed += 1

    rprint(f"Wrote {nr_passed} files to {output_dir} directory.")
    if stats_file:
        rprint(f"Statistics written to {stats_file.name}")


def _handle_filter_ss(args):
    input_dir = structure(args.input_dir, Path)
    output_dir = structure(args.output_dir, Path)
    copy_method: CopyMethod = structure(args.copy_method, CopyMethod)  # pyright: ignore[reportArgumentType]
    stats_file: TextIOWrapper | None = args.write_stats

    raw_query = {
        "abs_min_helix_residues": args.abs_min_helix_residues,
        "abs_max_helix_residues": args.abs_max_helix_residues,
        "abs_min_sheet_residues": args.abs_min_sheet_residues,
        "abs_max_sheet_residues": args.abs_max_sheet_residues,
        "ratio_min_helix_residues": args.ratio_min_helix_residues,
        "ratio_max_helix_residues": args.ratio_max_helix_residues,
        "ratio_min_sheet_residues": args.ratio_min_sheet_residues,
        "ratio_max_sheet_residues": args.ratio_max_sheet_residues,
    }
    query = converter.structure(raw_query, SecondaryStructureFilterQuery)
    input_files = sorted(glob_structure_files(input_dir))
    nr_total = len(input_files)
    output_dir.mkdir(parents=True, exist_ok=True)

    if stats_file:
        writer = csv.writer(stats_file)
        writer.writerow(
            [
                "input_file",
                "nr_residues",
                "nr_helix_residues",
                "nr_sheet_residues",
                "helix_ratio",
                "sheet_ratio",
                "passed",
                "output_file",
            ]
        )

    rprint(f"Filtering {nr_total} files in {input_dir} directory by secondary structure.")
    nr_passed = 0
    for input_file, result in filter_files_on_secondary_structure(input_files, query=query):
        output_file: Path | None = None
        if result.passed:
            output_file = output_dir / input_file.name
            copyfile(input_file, output_file, copy_method)
            nr_passed += 1
        if stats_file:
            writer.writerow(
                [
                    input_file,
                    result.stats.nr_residues,
                    result.stats.nr_helix_residues,
                    result.stats.nr_sheet_residues,
                    round(result.stats.helix_ratio, 3),
                    round(result.stats.sheet_ratio, 3),
                    result.passed,
                    output_file,
                ]
            )
    rprint(f"Wrote {nr_passed} files to {output_dir} directory.")
    if stats_file:
        rprint(f"Statistics written to {stats_file.name}")


def _handle_mcp(args):
    if find_spec("fastmcp") is None:
        msg = "Unable to start MCP server, please install `protein-quest[mcp]`."
        raise ImportError(msg)

    from protein_quest.mcp_server import mcp  # noqa: PLC0415

    if args.transport == "stdio":
        mcp.run(transport=args.transport)
    else:
        mcp.run(transport=args.transport, host=args.host, port=args.port)


HANDLERS: dict[tuple[str, str | None], Callable] = {
    ("search", "uniprot"): _handle_search_uniprot,
    ("search", "pdbe"): _handle_search_pdbe,
    ("search", "alphafold"): _handle_search_alphafold,
    ("search", "emdb"): _handle_search_emdb,
    ("search", "go"): _handle_search_go,
    ("search", "taxonomy"): _handle_search_taxonomy,
    ("search", "interaction-partners"): _handle_search_interaction_partners,
    ("search", "complexes"): _handle_search_complexes,
    ("retrieve", "pdbe"): _handle_retrieve_pdbe,
    ("retrieve", "alphafold"): _handle_retrieve_alphafold,
    ("retrieve", "emdb"): _handle_retrieve_emdb,
    ("filter", "confidence"): _handle_filter_confidence,
    ("filter", "chain"): _handle_filter_chain,
    ("filter", "residue"): _handle_filter_residue,
    ("filter", "secondary-structure"): _handle_filter_ss,
    ("mcp", None): _handle_mcp,
}


def _read_lines(file: TextIOWrapper) -> list[str]:
    return [line.strip() for line in file]


def _make_sure_parent_exists(file: TextIOWrapper):
    if file.name != "<stdout>":
        Path(file.name).parent.mkdir(parents=True, exist_ok=True)


def _write_lines(file: TextIOWrapper, lines: Iterable[str]):
    _make_sure_parent_exists(file)
    file.writelines(line + os.linesep for line in lines)


def _write_pdbe_csv(path: TextIOWrapper, data: dict[str, set[PdbResult]]):
    _make_sure_parent_exists(path)
    fieldnames = ["uniprot_acc", "pdb_id", "method", "resolution", "uniprot_chains", "chain"]
    writer = csv.DictWriter(path, fieldnames=fieldnames)
    writer.writeheader()
    for uniprot_acc, entries in sorted(data.items()):
        for e in sorted(entries, key=lambda x: (x.id, x.method)):
            writer.writerow(
                {
                    "uniprot_acc": uniprot_acc,
                    "pdb_id": e.id,
                    "method": e.method,
                    "resolution": e.resolution or "",
                    "uniprot_chains": e.uniprot_chains,
                    "chain": e.chain,
                }
            )


def _write_dict_of_sets2csv(file: TextIOWrapper, data: dict[str, set[str]], ref_id_field: str):
    _make_sure_parent_exists(file)
    fieldnames = ["uniprot_acc", ref_id_field]

    writer = csv.DictWriter(file, fieldnames=fieldnames)
    writer.writeheader()
    for uniprot_acc, ref_ids in sorted(data.items()):
        for ref_id in sorted(ref_ids):
            writer.writerow({"uniprot_acc": uniprot_acc, ref_id_field: ref_id})


def _iter_csv_rows(file: TextIOWrapper) -> Generator[dict[str, str]]:
    reader = csv.DictReader(file)
    yield from reader


def _read_column_from_csv(file: TextIOWrapper, column: str) -> set[str]:
    return {row[column] for row in _iter_csv_rows(file)}


def _write_complexes_csv(complexes: list[ComplexPortalEntry], output_csv: TextIOWrapper) -> None:
    """Write ComplexPortal information to a CSV file.

    Args:
        complexes: List of ComplexPortalEntry objects.
        output_csv: TextIOWrapper to write the CSV data to.
    """
    writer = csv.writer(output_csv)
    writer.writerow(
        [
            "query_protein",
            "complex_id",
            "complex_url",
            "complex_title",
            "members",
        ]
    )
    for entry in complexes:
        members_str = ";".join(sorted(entry.members))
        writer.writerow(
            [
                entry.query_protein,
                entry.complex_id,
                entry.complex_url,
                entry.complex_title,
                members_str,
            ]
        )



import pytest

from protein_quest.go import GoTerm, SearchResponse, converter, search_gene_ontology_term


def test_converter():
    input_data = b'{"numberOfHits":4,"results":[{"id":"GO:0006816","isObsolete":false,"name":"calcium ion transport","definition":{"text":"The directed movement of calcium (Ca) ions into, out of or within a cell, or between cells, by means of some agent such as a transporter or pore."},"aspect":"biological_process"},{"id":"GO:0000811","isObsolete":false,"name":"GINS complex","definition":{"text":"A heterotetrameric protein complex that associates with replication origins, where it is required for the initiation of DNA replication, and with replication forks."},"aspect":"cellular_component"},{"id":"GO:0070966","isObsolete":false,"name":"nuclear-transcribed mRNA catabolic process, no-go decay","definition":{"text":"The chemical reactions and pathways resulting in the breakdown of the transcript body of a nuclear-transcribed mRNA with stalls in translation elongation."},"aspect":"biological_process"},{"id":"GO:0001591","isObsolete":false,"name":"dopamine neurotransmitter receptor activity, coupled via Gi/Go","definition":{"text":"Combining with the neurotransmitter dopamine and activating adenylate cyclase via coupling to Gi/Go to initiate a change in cell activity."},"aspect":"molecular_function"}],"pageInfo":{"resultsPerPage":25,"current":1,"total":1}}'

    result = converter.loads(input_data, SearchResponse)

    assert isinstance(result, SearchResponse)
    assert len(result.results) == 4
    assert result.number_of_hits == 4
    assert (
        result.results[0].definition
        == "The directed movement of calcium (Ca) ions into, out of or within a cell, or between cells, by means of some agent such as a transporter or pore."
    )
    assert result.results[0].is_obsolete is False


@pytest.mark.vcr
@pytest.mark.asyncio
async def test_search_gene_ontology_term():
    results = await search_gene_ontology_term("GO:0043293", limit=1)

    expected = [
        GoTerm(
            id="GO:0043293",
            is_obsolete=False,
            name="apoptosome",
            definition="A multisubunit protein complex involved in the signaling "
            "phase of the apoptotic process. In mammals it is typically "
            "composed of seven Apaf-1 subunits bound to cytochrome c "
            "and caspase-9. A similar complex to promote apoptosis is "
            "formed from homologous gene products in other eukaryotic "
            "organisms.",
            aspect="cellular_component",
        )
    ]
    assert results == expected



import pytest

from protein_quest.taxonomy import Taxon, search_taxon


@pytest.fixture
def expected_human() -> Taxon:
    return Taxon(
        taxon_id="9606",
        scientific_name="Homo sapiens",
        common_name="Human",
        rank="species",
        other_names={
            "Homo sapiens Linnaeus, 1758",
            "human",
            "Home sapiens",
            "Homo sampiens",
            "Homo sapeins",
            "Homo sapian",
            "Homo sapians",
            "Homo sapien",
            "Homo sapience",
            "Homo sapiense",
            "Homo sapients",
            "Homo sapines",
            "Homo spaiens",
            "Homo spiens",
            "Humo sapiens",
            "Homo sapiens (SIRT6)",
            "Homo sapiens (PARIS)",
        },
    )


@pytest.mark.asyncio
@pytest.mark.vcr
async def test_search_taxon(expected_human: Taxon):
    results = await search_taxon("Human", limit=250)

    assert len(results) == 250
    expected0 = expected_human

    results0 = results[0]
    assert results0 == expected0


@pytest.mark.asyncio
@pytest.mark.vcr
async def test_search_taxon_by_id(expected_human: Taxon):
    results = await search_taxon("9606", field="tax_id")

    assert results == [expected_human]



"""Module for fetch Alphafold data."""

import logging
from asyncio import Semaphore
from collections.abc import AsyncGenerator, Iterable
from dataclasses import dataclass
from pathlib import Path
from typing import Literal, cast, get_args

from aiohttp_retry import RetryClient
from aiopath import AsyncPath
from tqdm.asyncio import tqdm
from yarl import URL

from protein_quest.alphafold.entry_summary import EntrySummary
from protein_quest.converter import converter
from protein_quest.utils import Cacher, PassthroughCacher, friendly_session, retrieve_files, run_async

logger = logging.getLogger(__name__)


DownloadableFormat = Literal[
    "summary",
    "bcif",
    "cif",
    "pdb",
    "paeImage",
    "paeDoc",
    "amAnnotations",
    "amAnnotationsHg19",
    "amAnnotationsHg38",
]
"""Types of formats that can be downloaded from the AlphaFold web service."""

downloadable_formats: set[DownloadableFormat] = set(get_args(DownloadableFormat))
"""Set of formats that can be downloaded from the AlphaFold web service."""


def _camel_to_snake_case(name: str) -> str:
    """Convert a camelCase string to snake_case."""
    return "".join(["_" + c.lower() if c.isupper() else c for c in name]).lstrip("_")


@dataclass
class AlphaFoldEntry:
    """AlphaFoldEntry represents a minimal single entry in the AlphaFold database.

    See https://alphafold.ebi.ac.uk/api-docs for more details on the API and data structure.
    """

    uniprot_acc: str
    summary: EntrySummary | None
    summary_file: Path | None = None
    bcif_file: Path | None = None
    cif_file: Path | None = None
    pdb_file: Path | None = None
    pae_image_file: Path | None = None
    pae_doc_file: Path | None = None
    am_annotations_file: Path | None = None
    am_annotations_hg19_file: Path | None = None
    am_annotations_hg38_file: Path | None = None

    @classmethod
    def format2attr(cls, dl_format: DownloadableFormat) -> str:
        """Get the attribute name for a specific download format.

        Args:
            dl_format: The format for which to get the attribute name.

        Returns:
            The attribute name corresponding to the download format.

        Raises:
            ValueError: If the format is not valid.
        """
        if dl_format not in downloadable_formats:
            msg = f"Invalid format: {dl_format}. Valid formats are: {downloadable_formats}"
            raise ValueError(msg)
        return _camel_to_snake_case(dl_format) + "_file"

    def by_format(self, dl_format: DownloadableFormat) -> Path | None:
        """Get the file path for a specific format.

        Args:
            dl_format: The format for which to get the file path.

        Returns:
            The file path corresponding to the download format.
            Or None if the file is not set.

        Raises:
            ValueError: If the format is not valid.
        """
        attr = self.format2attr(dl_format)
        return getattr(self, attr, None)

    def nr_of_files(self) -> int:
        """Nr of _file properties that are set

        Returns:
            The number of _file properties that are set.
        """
        return sum(1 for attr in vars(self) if attr.endswith("_file") and getattr(self, attr) is not None)


async def fetch_summary(
    qualifier: str, session: RetryClient, semaphore: Semaphore, save_dir: Path | None, cacher: Cacher
) -> list[EntrySummary]:
    """Fetches a summary from the AlphaFold database for a given qualifier.

    Args:
        qualifier: The uniprot accession for the protein or entry to fetch.
            For example `Q5VSL9`.
        session: An asynchronous HTTP client session with retry capabilities.
        semaphore: A semaphore to limit the number of concurrent requests.
        save_dir: An optional directory to save the fetched summary as a JSON file.
            If set and summary exists then summary will be loaded from disk instead of being fetched from the API.
            If not set then the summary will not be saved to disk and will always be fetched from the API.
        cacher: A cacher to use for caching the fetched summary. Only used if save_dir is not None.

    Returns:
        A list of EntrySummary objects representing the fetched summary.
    """
    url = f"https://alphafold.ebi.ac.uk/api/prediction/{qualifier}"
    fn: AsyncPath | None = None
    if save_dir is not None:
        fn = AsyncPath(save_dir / f"{qualifier}.json")
        cached_file = await cacher.copy_from_cache(Path(fn))
        if cached_file is not None:
            logger.debug(f"Using cached file {cached_file} for summary of {qualifier}.")
            raw_data = await AsyncPath(cached_file).read_bytes()
            return converter.loads(raw_data, list[EntrySummary])
        if await fn.exists():
            logger.debug(f"File {fn} already exists. Skipping download from {url}.")
            raw_data = await fn.read_bytes()
            return converter.loads(raw_data, list[EntrySummary])
    async with semaphore, session.get(url) as response:
        response.raise_for_status()
        raw_data = await response.content.read()
        if fn is not None:
            # TODO return fn and make it part of AlphaFoldEntry as summary_file prop
            await cacher.write_bytes(Path(fn), raw_data)
        return converter.loads(raw_data, list[EntrySummary])


async def fetch_summaries(
    qualifiers: Iterable[str],
    save_dir: Path | None = None,
    max_parallel_downloads: int = 5,
    cacher: Cacher | None = None,
) -> AsyncGenerator[EntrySummary]:
    semaphore = Semaphore(max_parallel_downloads)
    if save_dir is not None:
        save_dir.mkdir(parents=True, exist_ok=True)
    if cacher is None:
        cacher = PassthroughCacher()
    async with friendly_session() as session:
        tasks = [fetch_summary(qualifier, session, semaphore, save_dir, cacher) for qualifier in qualifiers]
        summaries_per_qualifier: list[list[EntrySummary]] = await tqdm.gather(
            *tasks, desc="Fetching Alphafold summaries"
        )
        for summaries in summaries_per_qualifier:
            for summary in summaries:
                yield summary


async def fetch_many_async(
    uniprot_accessions: Iterable[str],
    save_dir: Path,
    what: set[DownloadableFormat],
    max_parallel_downloads: int = 5,
    cacher: Cacher | None = None,
) -> AsyncGenerator[AlphaFoldEntry]:
    """Asynchronously fetches summaries and files from
    [AlphaFold Protein Structure Database](https://alphafold.ebi.ac.uk/).

    Args:
        uniprot_accessions: A set of Uniprot acessions to fetch.
        save_dir: The directory to save the fetched files to.
        what: A set of formats to download.
        max_parallel_downloads: The maximum number of parallel downloads.
        cacher: A cacher to use for caching the fetched files. Only used if summary is in what set.

    Yields:
        A dataclass containing the summary, pdb file, and pae file.
    """
    save_dir_for_summaries = save_dir if "summary" in what and save_dir is not None else None

    summaries = [
        s
        async for s in fetch_summaries(
            uniprot_accessions, save_dir_for_summaries, max_parallel_downloads=max_parallel_downloads, cacher=cacher
        )
    ]

    files = files_to_download(what, summaries)

    await retrieve_files(
        files,
        save_dir,
        desc="Downloading AlphaFold files",
        max_parallel_downloads=max_parallel_downloads,
        cacher=cacher,
    )
    for summary in summaries:
        yield AlphaFoldEntry(
            uniprot_acc=summary.uniprotAccession,
            summary=summary,
            summary_file=save_dir / f"{summary.uniprotAccession}.json" if save_dir_for_summaries is not None else None,
            bcif_file=save_dir / summary.bcifUrl.name if "bcif" in what else None,
            cif_file=save_dir / summary.cifUrl.name if "cif" in what else None,
            pdb_file=save_dir / summary.pdbUrl.name if "pdb" in what else None,
            pae_image_file=save_dir / summary.paeImageUrl.name if "paeImage" in what else None,
            pae_doc_file=save_dir / summary.paeDocUrl.name if "paeDoc" in what else None,
            am_annotations_file=(
                save_dir / summary.amAnnotationsUrl.name
                if "amAnnotations" in what and summary.amAnnotationsUrl
                else None
            ),
            am_annotations_hg19_file=(
                save_dir / summary.amAnnotationsHg19Url.name
                if "amAnnotationsHg19" in what and summary.amAnnotationsHg19Url
                else None
            ),
            am_annotations_hg38_file=(
                save_dir / summary.amAnnotationsHg38Url.name
                if "amAnnotationsHg38" in what and summary.amAnnotationsHg38Url
                else None
            ),
        )


def files_to_download(what: set[DownloadableFormat], summaries: Iterable[EntrySummary]) -> set[tuple[URL, str]]:
    if not (set(what) <= downloadable_formats):
        msg = (
            f"Invalid format(s) specified: {set(what) - downloadable_formats}. "
            f"Valid formats are: {downloadable_formats}"
        )
        raise ValueError(msg)

    files: set[tuple[URL, str]] = set()
    for summary in summaries:
        for fmt in what:
            if fmt == "summary":
                # summary is handled already in fetch_summary
                continue
            url = cast("URL | None", getattr(summary, f"{fmt}Url", None))
            if url is None:
                logger.warning(f"Summary {summary.uniprotAccession} does not have a URL for format '{fmt}'. Skipping.")
                continue
            file = (url, url.name)
            files.add(file)
    return files


def fetch_many(
    ids: Iterable[str],
    save_dir: Path,
    what: set[DownloadableFormat],
    max_parallel_downloads: int = 5,
    cacher: Cacher | None = None,
) -> list[AlphaFoldEntry]:
    """Synchronously fetches summaries and pdb and pae files from AlphaFold Protein Structure Database.

    Args:
        ids: A set of Uniprot IDs to fetch.
        save_dir: The directory to save the fetched files to.
        what: A set of formats to download.
        max_parallel_downloads: The maximum number of parallel downloads.
        cacher: A cacher to use for caching the fetched files. Only used if summary is in what set.

    Returns:
        A list of AlphaFoldEntry dataclasses containing the summary, pdb file, and pae file.
    """

    async def gather_entries():
        return [
            entry
            async for entry in fetch_many_async(
                ids, save_dir, what, max_parallel_downloads=max_parallel_downloads, cacher=cacher
            )
        ]

    return run_async(gather_entries())


def relative_to(entry: AlphaFoldEntry, session_dir: Path) -> AlphaFoldEntry:
    """Convert paths in an AlphaFoldEntry to be relative to the session directory.

    Args:
        entry: An AlphaFoldEntry instance with absolute paths.
        session_dir: The session directory to which the paths should be made relative.

    Returns:
        An AlphaFoldEntry instance with paths relative to the session directory.
    """
    return AlphaFoldEntry(
        uniprot_acc=entry.uniprot_acc,
        summary=entry.summary,
        summary_file=entry.summary_file.relative_to(session_dir) if entry.summary_file else None,
        bcif_file=entry.bcif_file.relative_to(session_dir) if entry.bcif_file else None,
        cif_file=entry.cif_file.relative_to(session_dir) if entry.cif_file else None,
        pdb_file=entry.pdb_file.relative_to(session_dir) if entry.pdb_file else None,
        pae_image_file=entry.pae_image_file.relative_to(session_dir) if entry.pae_image_file else None,
        pae_doc_file=entry.pae_doc_file.relative_to(session_dir) if entry.pae_doc_file else None,
        am_annotations_file=entry.am_annotations_file.relative_to(session_dir) if entry.am_annotations_file else None,
        am_annotations_hg19_file=(
            entry.am_annotations_hg19_file.relative_to(session_dir) if entry.am_annotations_hg19_file else None
        ),
        am_annotations_hg38_file=(
            entry.am_annotations_hg38_file.relative_to(session_dir) if entry.am_annotations_hg38_file else None
        ),
    )



"""Module for Gene Ontology (GO) functions."""

import csv
import logging
from collections.abc import Generator
from dataclasses import dataclass
from io import TextIOWrapper
from typing import Literal, get_args

from cattrs.gen import make_dict_structure_fn, override

from protein_quest.converter import converter
from protein_quest.utils import friendly_session

logger = logging.getLogger(__name__)

Aspect = Literal["cellular_component", "biological_process", "molecular_function"]
"""The aspect of the GO term."""
allowed_aspects = set(get_args(Aspect))
"""Allowed aspects for GO terms."""


@dataclass(frozen=True, slots=True)
class GoTerm:
    """A Gene Ontology (GO) term.

    Parameters:
        id: The unique identifier for the GO term, e.g., 'GO:0043293'.
        is_obsolete: Whether the GO term is obsolete.
        name: The name of the GO term.
        definition: The definition of the GO term.
        aspect: The aspect of the GO term.
    """

    id: str
    is_obsolete: bool
    name: str
    definition: str
    aspect: Aspect


@dataclass(frozen=True, slots=True)
class PageInfo:
    current: int
    total: int


@dataclass(frozen=True, slots=True)
class SearchResponse:
    results: list[GoTerm]
    number_of_hits: int
    page_info: PageInfo


def flatten_definition(definition, _context) -> str:
    return definition["text"]


# Use hook to convert incoming camelCase to snake_case
# and to flatten definition {text} to text
# see https://catt.rs/en/stable/customizing.html#rename
converter.register_structure_hook(
    GoTerm,
    make_dict_structure_fn(
        GoTerm,
        converter,
        is_obsolete=override(rename="isObsolete"),
        definition=override(struct_hook=flatten_definition),
    ),
)
converter.register_structure_hook(
    SearchResponse,
    make_dict_structure_fn(
        SearchResponse, converter, number_of_hits=override(rename="numberOfHits"), page_info=override(rename="pageInfo")
    ),
)


async def search_gene_ontology_term(
    term: str, aspect: Aspect | None = None, include_obsolete: bool = False, limit: int = 100
) -> list[GoTerm]:
    """Search for a Gene Ontology (GO) term by its name or ID.

    Calls the EBI QuickGO API at https://www.ebi.ac.uk/QuickGO/api/index.html .

    Examples:
        To search for `apoptosome` terms do.

        >>> from protein_quest.go import search_go_term
        >>> r = await search_go_term('apoptosome')
        >>> len(r)
        5
        >>> r[0]
        GoTerm(id='GO:0043293', is_obsolete=False, name='apoptosome', definition='A multisubunit protein ...')

    Args:
        term: The GO term to search for. For example `nucleus` or `GO:0006816`.
        aspect: The aspect to filter by. If not given, all aspects are included.
        include_obsolete: Whether to include obsolete terms. By default, obsolete terms are excluded.
        limit: The maximum number of results to return.

    Returns:
        List of GO terms

    Raises:
        ValueError: If the aspect is invalid.
    """
    url = "https://www.ebi.ac.uk/QuickGO/services/ontology/go/search"
    page_limit = 100
    params = {"query": term, "limit": str(page_limit), "page": "1"}
    if aspect is not None and aspect not in allowed_aspects:
        msg = f"Invalid aspect: {aspect}. Allowed aspects are: {allowed_aspects} or None."
        raise ValueError(msg)
    logger.debug("Fetching GO terms from %s with params %s", url, params)
    async with friendly_session() as session:
        # Fetch first page to learn how many pages there are
        async with session.get(url, params=params) as response:
            response.raise_for_status()
            raw_data = await response.read()
            data = converter.loads(raw_data, SearchResponse)

        terms = list(_filter_go_terms(data.results, aspect, include_obsolete))
        if len(terms) >= limit:
            # Do not fetch additional pages if we have enough results
            return terms[:limit]
        total_pages = data.page_info.total
        logger.debug("GO search returned %s pages (current=%s)", total_pages, data.page_info.current)

        # Retrieve remaining pages (if any) and extend results
        if total_pages > 1:
            for page in range(2, total_pages + 1):
                params["page"] = str(page)
                logger.debug("Fetching additional GO terms page %s/%s with params %s", page, total_pages, params)
                async with session.get(url, params=params) as response:
                    response.raise_for_status()
                    raw_data = await response.read()
                    data = converter.loads(raw_data, SearchResponse)
                terms.extend(_filter_go_terms(data.results, aspect, include_obsolete))
                if len(terms) >= limit:
                    # Do not fetch additional pages if we have enough results
                    break

        return terms[:limit]


def _filter_go_terms(terms: list[GoTerm], aspect: Aspect | None, include_obsolete: bool) -> Generator[GoTerm]:
    for oboterm in terms:
        if not include_obsolete and oboterm.is_obsolete:
            continue
        if aspect and oboterm.aspect != aspect:
            continue
        yield oboterm


def write_go_terms_to_csv(terms: list[GoTerm], csv_file: TextIOWrapper) -> None:
    """Write a list of GO terms to a CSV file.

    Args:
        terms: The list of GO terms to write.
        csv_file: The CSV file to write to.
    """
    writer = csv.writer(csv_file)
    writer.writerow(["id", "name", "aspect", "definition"])
    for term in terms:
        writer.writerow([term.id, term.name, term.aspect, term.definition])



import asyncio
from pathlib import Path

import pytest
from _pytest.capture import CaptureFixture
from aiohttp.streams import AsyncStreamIterator

from protein_quest.utils import (
    CopyMethod,
    DirectoryCacher,
    PassthroughCacher,
    async_copyfile,
    copyfile,
    populate_cache_command,
    user_cache_root_dir,
)


def test_copyfile_copy(tmp_path: Path):
    src = tmp_path / "src.txt"
    dst = tmp_path / "dst.txt"
    src.write_text("Hello, World!")
    copyfile(src, dst, "copy")
    assert dst.read_text() == "Hello, World!"
    assert not dst.is_symlink()


def test_copyfile_symlink(tmp_path: Path):
    src = tmp_path / "src.txt"
    dst = tmp_path / "dst.txt"
    src.write_text("Hello, World!")
    copyfile(src, dst, "symlink")
    assert dst.read_text() == "Hello, World!"
    assert dst.is_symlink()
    assert dst.resolve() == src


def test_copyfile_hardlink(tmp_path: Path):
    src = tmp_path / "src.txt"
    dst = tmp_path / "dst.txt"
    src.write_text("Hello, World!")
    copyfile(src, dst, "hardlink")
    assert dst.read_text() == "Hello, World!"
    assert not dst.is_symlink()
    assert src.stat().st_ino == dst.stat().st_ino
    assert src.stat().st_nlink == 2
    assert dst.stat().st_nlink == 2


def test_copyfile_invalid_method(tmp_path: Path):
    src = tmp_path / "src.txt"
    dst = tmp_path / "dst.txt"
    with pytest.raises(ValueError, match="Unknown method"):
        copyfile(src, dst, "invalid")  # type: ignore  # noqa: PGH003


@pytest.mark.asyncio
async def test_async_copyfile_copy(tmp_path: Path):
    src = tmp_path / "src.txt"
    dst = tmp_path / "dst.txt"
    src.write_text("Hello, World!")
    await async_copyfile(src, dst, "copy")
    assert dst.read_text() == "Hello, World!"
    assert not dst.is_symlink()


@pytest.mark.asyncio
async def test_async_copyfile_symlink(tmp_path: Path):
    src = tmp_path / "src.txt"
    dst = tmp_path / "dst.txt"
    src.write_text("Hello, World!")
    await async_copyfile(src, dst, "symlink")
    assert dst.read_text() == "Hello, World!"
    assert dst.is_symlink()
    assert dst.resolve() == src


@pytest.mark.asyncio
async def test_async_copyfile_hardlink(tmp_path: Path):
    src = tmp_path / "src.txt"
    dst = tmp_path / "dst.txt"
    src.write_text("Hello, World!")
    await async_copyfile(src, dst, "hardlink")
    assert dst.read_text() == "Hello, World!"
    assert not dst.is_symlink()
    assert src.stat().st_ino == dst.stat().st_ino
    assert src.stat().st_nlink == 2
    assert dst.stat().st_nlink == 2


@pytest.mark.asyncio
async def test_async_copyfile_invalid_method(tmp_path: Path):
    src = tmp_path / "src.txt"
    dst = tmp_path / "dst.txt"
    with pytest.raises(ValueError, match="Unknown method"):
        await async_copyfile(src, dst, "invalid")  # type: ignore  # noqa: PGH003


def test_user_cache_root_dir():
    cache_dir = user_cache_root_dir()
    assert cache_dir.name == "protein-quest"


class ByteGenerator(AsyncStreamIterator[bytes]):
    """Mock for return of aiohttp.ClientResponse.content.iter_chunked()"""

    def __init__(self, data: bytes, chunk_size: int = 5):
        self.data = data
        self.chunk_size = chunk_size

    def __aiter__(self):
        self.index = 0
        return self

    async def __anext__(self):
        if self.index >= len(self.data):
            raise StopAsyncIteration
        chunk = self.data[self.index : self.index + self.chunk_size]
        self.index += self.chunk_size
        await asyncio.sleep(0)  # Yield control to the event loop
        return chunk


def test_DirectoryCacher_init_copy_method_invalid(tmp_path: Path):
    with pytest.raises(ValueError, match="Unknown copy method"):
        DirectoryCacher(tmp_path / "cache", copy_method="invalid")  # type: ignore  # noqa: PGH003


@pytest.mark.parametrize("copy_method", ["copy", "symlink", "hardlink"])
class TestDirectoryCacher:
    @pytest.fixture
    def cacher(self, tmp_path: Path, copy_method: CopyMethod) -> DirectoryCacher:
        return DirectoryCacher(tmp_path / "cache", copy_method=copy_method)

    def test_init(self, tmp_path: Path, cacher: DirectoryCacher):
        assert cacher.cache_dir == tmp_path / "cache"
        assert cacher.cache_dir.exists()
        assert cacher.cache_dir.is_dir()

    @pytest.mark.asyncio
    async def test_write_bytes(self, tmp_path: Path, cacher: DirectoryCacher):
        target = tmp_path / "test.txt"

        cache_file = await cacher.write_bytes(target, b"Hello, World!")

        assert cache_file.exists()
        assert cache_file.read_bytes() == b"Hello, World!"

    def test_in_missing(self, cacher: DirectoryCacher):
        assert "any_file.txt" not in cacher

    @pytest.mark.asyncio
    async def test_in_present(self, tmp_path: Path, cacher: DirectoryCacher):
        # Fill cache
        target = tmp_path / "test.txt"
        await cacher.write_bytes(target, b"Hello, World!")

        assert target in cacher

    @pytest.mark.asyncio
    async def test_copy_from_cache_missing(self, tmp_path: Path, cacher: DirectoryCacher):
        target = tmp_path / "test.txt"

        result = await cacher.copy_from_cache(target)
        assert result is None

    @pytest.mark.asyncio
    async def test_copy_from_cache_present(self, tmp_path: Path, cacher: DirectoryCacher):
        other = tmp_path / "other"
        other.mkdir()
        source = other / "test.txt"
        await cacher.write_bytes(source, b"Hello, World!")

        target = tmp_path / "test.txt"
        result = await cacher.copy_from_cache(target)

        assert result is not None
        assert result.exists()
        assert result.read_bytes() == b"Hello, World!"

    @pytest.mark.asyncio
    async def test_write_iter(self, tmp_path: Path, cacher: DirectoryCacher):
        target = tmp_path / "test.txt"

        cache_file = await cacher.write_iter(target, ByteGenerator(b"Hello, World!"))

        assert cache_file.exists()
        assert cache_file.read_bytes() == b"Hello, World!"

    @pytest.mark.asyncio
    async def test_write_bytes_twice(self, tmp_path: Path, cacher: DirectoryCacher):
        target = tmp_path / "test.txt"

        # First write
        cache_file = await cacher.write_bytes(target, b"Hello, World!")
        assert cache_file.exists()
        assert cache_file.read_bytes() == b"Hello, World!"

        # Second write
        with pytest.raises(FileExistsError, match=str(cache_file)):
            await cacher.write_bytes(target, b"Goodbye, World!")

    @pytest.mark.asyncio
    async def test_write_iter_twice(self, tmp_path: Path, cacher: DirectoryCacher):
        target = tmp_path / "test.txt"

        # First write
        cache_file = await cacher.write_iter(target, ByteGenerator(b"Hello, World!"))
        assert cache_file.exists()
        assert cache_file.read_bytes() == b"Hello, World!"

        # Second write
        with pytest.raises(FileExistsError, match=str(cache_file)):
            await cacher.write_iter(target, ByteGenerator(b"Goodbye, World!"))


class TestPassthroughCacher:
    @pytest.mark.asyncio
    async def test_write_bytes(self, tmp_path: Path):
        cacher = PassthroughCacher()
        target = tmp_path / "test.txt"

        cache_file = await cacher.write_bytes(target, b"Hello, World!")

        assert cache_file.exists()
        assert cache_file.read_bytes() == b"Hello, World!"

    def test_in_missing(self, tmp_path: Path):
        cacher = PassthroughCacher()
        assert "any_file.txt" not in cacher

    @pytest.mark.asyncio
    async def test_in_present(self, tmp_path: Path):
        # Fill cache
        cacher = PassthroughCacher()
        target = tmp_path / "test.txt"
        await cacher.write_bytes(target, b"Hello, World!")

        # PassthroughCacher never has anything cached
        assert target not in cacher

    @pytest.mark.asyncio
    async def test_copy_from_cache_missing(self, tmp_path: Path):
        cacher = PassthroughCacher()
        target = tmp_path / "test.txt"

        result = await cacher.copy_from_cache(target)
        assert result is None

    @pytest.mark.asyncio
    async def test_copy_from_cache_present(self, tmp_path: Path):
        cacher = PassthroughCacher()
        target = tmp_path / "test.txt"
        await cacher.write_bytes(target, b"Hello, World!")

        result = await cacher.copy_from_cache(target)

        # PassthroughCacher never has anything cached
        assert result is None

    @pytest.mark.asyncio
    async def test_write_iter(self, tmp_path: Path):
        cacher = PassthroughCacher()
        target = tmp_path / "test.txt"

        cache_file = await cacher.write_iter(target, ByteGenerator(b"Hello, World!"))

        assert cache_file.exists()
        assert cache_file.read_bytes() == b"Hello, World!"

    @pytest.mark.asyncio
    async def test_write_bytes_twice(self, tmp_path: Path):
        cacher = PassthroughCacher()
        target = tmp_path / "test.txt"

        # First write
        cache_file = await cacher.write_bytes(target, b"Hello, World!")
        assert cache_file.exists()
        assert cache_file.read_bytes() == b"Hello, World!"

        # Second write
        with pytest.raises(FileExistsError, match=str(target)):
            await cacher.write_bytes(target, b"Goodbye, World!")

    @pytest.mark.asyncio
    async def test_write_iter_twice(self, tmp_path: Path):
        cacher = PassthroughCacher()
        target = tmp_path / "test.txt"

        # First write
        cache_file = await cacher.write_iter(target, ByteGenerator(b"Hello, World!"))
        assert cache_file.exists()
        assert cache_file.read_bytes() == b"Hello, World!"

        # Second write
        with pytest.raises(FileExistsError, match=str(target)):
            await cacher.write_iter(target, ByteGenerator(b"Goodbye, World!"))


def test_populate_cache_command_with_hardlink(tmp_path: Path, capsys: CaptureFixture[str]):
    source_dir = tmp_path / "source"
    source_dir.mkdir()
    src1 = source_dir / "file1.txt"
    src2 = source_dir / "file2.txt"
    src1.write_text("File 1")
    src2.write_text("File 2")

    cache_dir = tmp_path / "cache"
    cache_dir.mkdir()

    populate_cache_command(
        [
            "populate-cache",
            str(source_dir),
            "--cache-dir",
            str(cache_dir),
            "--copy-method",
            "hardlink",
        ]
    )

    assert src1.stat().st_nlink == 2
    assert src2.stat().st_nlink == 2

    captured = capsys.readouterr()
    assert "->" in captured.out



"""Dask helper functions."""

import logging
import os
from collections.abc import Callable, Collection
from typing import Concatenate, ParamSpec, cast

from dask.distributed import Client, LocalCluster, progress
from distributed.deploy.cluster import Cluster
from psutil import cpu_count

logger = logging.getLogger(__name__)


def configure_dask_scheduler(
    scheduler_address: str | Cluster | None,
    name: str,
    nproc: int = 1,
) -> str | Cluster:
    """Configure the Dask scheduler by reusing existing or creating a new cluster.

    Args:
        scheduler_address: Address of the Dask scheduler to connect to, or None for local cluster.
        name: Name for the Dask cluster.
        nproc: Number of processes to use per worker for CPU support.

    Returns:
        A Dask Cluster instance or a string address for the scheduler.
    """
    if scheduler_address is None:
        scheduler_address = _configure_cpu_dask_scheduler(nproc, name)
        logger.info(f"Using local Dask cluster: {scheduler_address}")

    return scheduler_address


def nr_cpus() -> int:
    """Determine the number of CPU cores to use.

    If the environment variables SLURM_CPUS_PER_TASK or OMP_NUM_THREADS are set,
    their value is used. Otherwise, the number of physical CPU cores is returned.

    Returns:
        The number of CPU cores to use.

    Raises:
        ValueError: If the number of physical CPU cores cannot be determined.
    """
    physical_cores = cpu_count(logical=False)
    if physical_cores is None:
        msg = "Cannot determine number of logical CPU cores."
        raise ValueError(msg)
    for var in ["SLURM_CPUS_PER_TASK", "OMP_NUM_THREADS"]:
        value = os.environ.get(var)
        if value is not None:
            logger.warning(
                'Not using all CPU cores (%s) of machine, environment variable "%s" is set to %s.',
                physical_cores,
                var,
                value,
            )
            return int(value)
    return physical_cores


def _configure_cpu_dask_scheduler(nproc: int, name: str) -> LocalCluster:
    total_cpus = nr_cpus()
    n_workers = total_cpus // nproc
    # Use single thread per worker to prevent GIL slowing down the computations
    return LocalCluster(name=name, threads_per_worker=1, n_workers=n_workers)


# Generic type parameters used across helpers
P = ParamSpec("P")


def dask_map_with_progress[T, R, **P](
    client: Client,
    func: Callable[Concatenate[T, P], R],
    iterable: Collection[T],
    *args: P.args,
    **kwargs: P.kwargs,
) -> list[R]:
    """
    Wrapper for map, progress, and gather of Dask that returns a correctly typed list.

    Args:
        client: Dask client.
        func: Function to map; first parameter comes from ``iterable`` and any
            additional parameters can be provided positionally via ``*args`` or
            as keyword arguments via ``**kwargs``.
        iterable: Collection of arguments to map over.
        *args: Additional positional arguments to pass to client.map().
        **kwargs: Additional keyword arguments to pass to client.map().

    Returns:
        List of results of type returned by `func` function.
    """
    if client.dashboard_link:
        logger.info(f"Follow progress on dask dashboard at: {client.dashboard_link}")
    futures = client.map(func, iterable, *args, **kwargs)
    progress(futures)
    results = client.gather(futures)
    return cast("list[R]", results)



"""Convert json or dict to Python objects."""

from cattrs.preconf.orjson import make_converter
from yarl import URL

type Percentage = float
"""Type alias for percentage values (0.0-100.0)."""
type Ratio = float
"""Type alias for ratio values (0.0-1.0)."""
type PositiveInt = int
"""Type alias for positive integer values (>= 0)."""

converter = make_converter()
"""cattrs converter to read JSON document or dict to Python objects."""
converter.register_structure_hook(URL, lambda v, _: URL(v))
converter.register_unstructure_hook(URL, lambda u: str(u))


@converter.register_structure_hook
def percentage_hook(val, _) -> Percentage:
    value = float(val)
    """Cattrs hook to validate percentage values."""
    if not 0.0 <= value <= 100.0:
        msg = f"Value {value} is not a valid percentage (0.0-100.0)"
        raise ValueError(msg)
    return value


@converter.register_structure_hook
def ratio_hook(val, _) -> Ratio:
    """Cattrs hook to validate ratio values."""
    value = float(val)
    if not 0.0 <= value <= 1.0:
        msg = f"Value {value} is not a valid ratio (0.0-1.0)"
        raise ValueError(msg)
    return value


@converter.register_structure_hook
def positive_int_hook(val, _) -> PositiveInt:
    """Cattrs hook to validate positive integer values."""
    value = int(val)
    if value < 0:
        msg = f"Value {value} is not a valid positive integer (>= 0)"
        raise ValueError(msg)
    return value



from pathlib import Path

import gemmi
import pytest

from protein_quest.converter import converter
from protein_quest.ss import (
    SecondaryStructureFilterQuery,
    SecondaryStructureFilterResult,
    SecondaryStructureStats,
    filter_file_on_secondary_structure,
    filter_files_on_secondary_structure,
    filter_on_secondary_structure,
    nr_of_residues_in_helix,
    nr_of_residues_in_sheet,
    nr_of_residues_in_total,
)


@pytest.fixture
def sample_cif() -> Path:
    """Downloaded from https://www.rcsb.org/structure/3JRS
    and filtered with
    `write_single_chain_pdb_file(Path('tests/fixtures/3JRS.cif.gz'), 'B', Path('tests/fixtures/'))`
    """
    return Path(__file__).parent / "fixtures" / "3JRS_B2A.cif.gz"


@pytest.fixture
def sample_structure(sample_cif: Path) -> gemmi.Structure:
    return gemmi.read_structure(str(sample_cif))


@pytest.fixture(scope="module")
def sample_stats() -> SecondaryStructureStats:
    return SecondaryStructureStats(
        nr_residues=173,
        nr_helix_residues=58,
        nr_sheet_residues=59,
        helix_ratio=58 / 173,
        sheet_ratio=59 / 173,
    )


def test_nr_of_residues_in_helix(sample_structure: gemmi.Structure):
    expected_count = 58
    assert nr_of_residues_in_helix(sample_structure) == expected_count


def test_nr_of_residues_in_sheet(sample_structure: gemmi.Structure):
    expected_count = 59
    assert nr_of_residues_in_sheet(sample_structure) == expected_count


def test_nr_of_residues_in_total(sample_structure: gemmi.Structure):
    expected_count = 173
    assert nr_of_residues_in_total(sample_structure) == expected_count


@pytest.mark.parametrize(
    "query, expected_passed",
    [
        # abs_min_helix_residues
        (SecondaryStructureFilterQuery(abs_min_helix_residues=58), True),
        (SecondaryStructureFilterQuery(abs_min_helix_residues=59), False),
        # abs_max_helix_residues
        (SecondaryStructureFilterQuery(abs_max_helix_residues=58), True),
        (SecondaryStructureFilterQuery(abs_max_helix_residues=57), False),
        # abs_min_sheet_residues
        (SecondaryStructureFilterQuery(abs_min_sheet_residues=59), True),
        (SecondaryStructureFilterQuery(abs_min_sheet_residues=60), False),
        # abs_max_sheet_residues
        (SecondaryStructureFilterQuery(abs_max_sheet_residues=59), True),
        (SecondaryStructureFilterQuery(abs_max_sheet_residues=58), False),
        # ratio_min_helix_residues
        (SecondaryStructureFilterQuery(ratio_min_helix_residues=58 / 173), True),
        (SecondaryStructureFilterQuery(ratio_min_helix_residues=0.4), False),
        # ratio_max_helix_residues
        (SecondaryStructureFilterQuery(ratio_max_helix_residues=58 / 173), True),
        (SecondaryStructureFilterQuery(ratio_max_helix_residues=0.3), False),
        # ratio_min_sheet_residues
        (SecondaryStructureFilterQuery(ratio_min_sheet_residues=59 / 173), True),
        (SecondaryStructureFilterQuery(ratio_min_sheet_residues=0.4), False),
        # ratio_max_sheet_residues
        (SecondaryStructureFilterQuery(ratio_max_sheet_residues=59 / 173), True),
        (SecondaryStructureFilterQuery(ratio_max_sheet_residues=0.3), False),
        # multiple
        (
            SecondaryStructureFilterQuery(
                ratio_min_helix_residues=0.1,
                ratio_min_sheet_residues=0.1,
            ),
            True,
        ),
        (
            SecondaryStructureFilterQuery(
                ratio_min_helix_residues=0.9,
                ratio_min_sheet_residues=0.9,
            ),
            False,
        ),
    ],
)
def test_filter_on_secondary_structure(
    sample_structure: gemmi.Structure,
    sample_stats: SecondaryStructureStats,
    query: SecondaryStructureFilterQuery,
    expected_passed: bool,
):
    assert query.is_actionable()
    result = filter_on_secondary_structure(sample_structure, query)
    expected = SecondaryStructureFilterResult(
        stats=sample_stats,
        passed=expected_passed,
    )
    assert result == expected


def test_SecondaryStructureFilterQuery_non_actionable():
    query = SecondaryStructureFilterQuery()
    assert not query.is_actionable()


def test_filter_on_secondary_structure_raises_on_zero_conditions(sample_structure: gemmi.Structure):
    query = SecondaryStructureFilterQuery()
    with pytest.raises(ValueError, match="No filtering conditions provided"):
        filter_on_secondary_structure(sample_structure, query)


def test_filter_on_secondary_structure_raise_on_zero_residues():
    structure = gemmi.Structure()
    query = SecondaryStructureFilterQuery(abs_min_helix_residues=1)
    with pytest.raises(ValueError, match="Structure has zero residues"):
        filter_on_secondary_structure(structure, query)


def test_filter_file_on_secondary_structure(sample_cif: Path, sample_stats: SecondaryStructureStats):
    query = SecondaryStructureFilterQuery(abs_min_helix_residues=1)
    result = filter_file_on_secondary_structure(sample_cif, query)
    expected = SecondaryStructureFilterResult(stats=sample_stats, passed=True)
    assert result == expected


def test_filter_files_on_secondary_structure(sample_cif: Path, sample_stats: SecondaryStructureStats):
    query = SecondaryStructureFilterQuery(abs_min_helix_residues=1)
    result = filter_files_on_secondary_structure([sample_cif], query)
    expected = {sample_cif: SecondaryStructureFilterResult(stats=sample_stats, passed=True)}
    assert dict(result) == expected


def test_converter():
    raw = {
        "abs_min_helix_residues": 10,
        "abs_max_helix_residues": 20,
        "abs_min_sheet_residues": 5,
        "abs_max_sheet_residues": 15,
        "ratio_min_helix_residues": 0.1,
        "ratio_max_helix_residues": 0.3,
        "ratio_min_sheet_residues": 0.2,
        "ratio_max_sheet_residues": 0.4,
    }
    result = converter.structure(raw, SecondaryStructureFilterQuery)
    expected = SecondaryStructureFilterQuery(
        abs_min_helix_residues=10,
        abs_max_helix_residues=20,
        abs_min_sheet_residues=5,
        abs_max_sheet_residues=15,
        ratio_min_helix_residues=0.1,
        ratio_max_helix_residues=0.3,
        ratio_min_sheet_residues=0.2,
        ratio_max_sheet_residues=0.4,
    )
    assert result == expected


@pytest.mark.parametrize(
    "raw, match",
    [
        (
            {
                "abs_min_helix_residues": 20,
                "abs_max_helix_residues": 10,
            },
            "must be smaller than max",
        ),
        (
            {
                "abs_min_sheet_residues": 15,
                "abs_max_sheet_residues": 5,
            },
            "must be smaller than max",
        ),
        (
            {
                "ratio_min_helix_residues": 0.4,
                "ratio_max_helix_residues": 0.3,
            },
            "must be smaller than max",
        ),
        (
            {
                "ratio_min_sheet_residues": 0.5,
                "ratio_max_sheet_residues": 0.4,
            },
            "must be smaller than max",
        ),
    ],
)
def test_converter_raises_on_invalid_range(raw, match):
    with pytest.raises(ValueError, match=match):
        converter.structure(raw, SecondaryStructureFilterQuery)


@pytest.mark.parametrize(
    "raw, match",
    [
        (
            {
                "ratio_min_sheet_residues": 4.2,
            },
            "is not a valid ratio",
        ),
        (
            {
                "ratio_min_sheet_residues": "not a number",
            },
            "could not convert string to float",
        ),
    ],
)
def test_converter_raises_on_invalid_ratio(raw, match):
    with pytest.RaisesGroup(pytest.RaisesExc(ValueError, match=match)):
        converter.structure(raw, SecondaryStructureFilterQuery)



"""MCP server for protein-quest.

Can be run with:

```shell
# for development
fastmcp dev src/protein_quest/mcp_server.py
# or from inspector
npx @modelcontextprotocol/inspector
# tranport type: stdio
# comand: protein-quest
# arguments: mcp

# or with server and inspector
protein-quest mcp --transport streamable-http
# in another shell
npx @modelcontextprotocol/inspector
# transport type: streamable http
# URL: http://127.0.0.1:8000/mcp

# or with copilot in VS code
# ctrl + shift + p
# mcp: add server...
# Choose STDIO
# command: uv run protein-quest mcp
# id: protein-quest
```

Examples:

   - What are the PDBe structures for `A8MT69` uniprot accession?

"""

from collections.abc import Mapping
from pathlib import Path
from textwrap import dedent
from typing import Annotated

from fastmcp import FastMCP
from pydantic import Field

from protein_quest.alphafold.confidence import ConfidenceFilterQuery, ConfidenceFilterResult, filter_file_on_residues
from protein_quest.alphafold.fetch import AlphaFoldEntry, DownloadableFormat
from protein_quest.alphafold.fetch import fetch_many as alphafold_fetch
from protein_quest.emdb import fetch as emdb_fetch
from protein_quest.go import search_gene_ontology_term
from protein_quest.pdbe.fetch import fetch as pdbe_fetch
from protein_quest.pdbe.io import glob_structure_files, nr_residues_in_chain, write_single_chain_pdb_file
from protein_quest.ss import filter_file_on_secondary_structure
from protein_quest.taxonomy import search_taxon
from protein_quest.uniprot import (
    PdbResult,
    Query,
    search4af,
    search4emdb,
    search4macromolecular_complexes,
    search4pdb,
    search4uniprot,
)

mcp = FastMCP("protein-quest")

# do not want to make dataclasses in non-mcp code into Pydantic models,
# so we use Annotated here to add description on roots.


@mcp.tool
def search_uniprot(
    uniprot_query: Annotated[Query, Field(description=Query.__doc__)],
    limit: Annotated[int, Field(gt=0, description="Limit the number of uniprot accessions returned")] = 100,
) -> set[str]:
    """Search UniProt for proteins matching the given query."""
    return search4uniprot(uniprot_query, limit=limit)


@mcp.tool
def search_pdb(
    uniprot_accs: set[str],
    limit: Annotated[int, Field(gt=0, description="Limit the number of entries returned")] = 100,
) -> Annotated[
    dict[str, set[PdbResult]],
    Field(
        description=dedent(f"""\
            Dictionary with protein IDs as keys and sets of PDB results as values.
            A PDB result is {PdbResult.__doc__}""")
    ),
]:
    """Search PDBe structures for given uniprot accessions."""
    return search4pdb(uniprot_accs, limit=limit)


@mcp.tool
async def fetch_pdbe_structures(pdb_ids: set[str], save_dir: Path) -> Mapping[str, Path]:
    """Fetch the PDBe structures for given PDB IDs.

    Args:
        pdb_ids: A set of PDB IDs.
        save_dir: The directory to save the fetched files.

    Returns:
        A mapping of PDB ID to the path of the fetched structure file.
    """
    return await pdbe_fetch(pdb_ids, save_dir)


@mcp.tool
def extract_single_chain_from_structure(
    input_file: Path,
    chain2keep: str,
    output_dir: Path,
    out_chain: str = "A",
) -> Path:
    """
    Extract a single chain from a mmCIF/pdb file and write to a new file.

    Args:
        input_file: Path to the input mmCIF/pdb file.
        chain2keep: The chain to keep.
        output_dir: Directory to save the output file.
        out_chain: The chain identifier for the output file.

    Returns:
        Path to the output mmCIF/pdb file
    """
    return write_single_chain_pdb_file(input_file, chain2keep, output_dir, out_chain)


@mcp.tool
def list_structure_files(path: Path) -> list[Path]:
    """List structure files (.pdb, .pdb.gz, .cif, .cif.gz) in the specified directory."""
    return list(glob_structure_files(path))


# TODO replace remaining decorators with wrapper if tool does single function call
# so we do not have to replicate docstring,
# minor con is that it does not show up in api docs
mcp.tool(nr_residues_in_chain)
mcp.tool(search_taxon)
mcp.tool(search_gene_ontology_term)


@mcp.tool
def search_alphafolds(
    uniprot_accs: set[str],
    limit: Annotated[int, Field(gt=0, description="Limit the number of entries returned")] = 100,
) -> Annotated[
    set[str],
    Field(description="Set of uniprot accessions which have an AlphaFold entry"),
]:
    """Search for AlphaFold entries in UniProtKB accessions."""
    # each uniprot accesion can have one or more AlphaFold IDs
    # an AlphaFold ID is the same as the uniprot accession
    # so we return a subset of uniprot_accs
    results = search4af(uniprot_accs, limit)
    return {k for k, v in results.items() if v}


mcp.tool(search4emdb, name="search_emdb")
mcp.tool(search4macromolecular_complexes, name="search_macromolecular_complexes")


@mcp.tool
def fetch_alphafold_structures(uniprot_accs: set[str], save_dir: Path) -> list[AlphaFoldEntry]:
    """Fetch the AlphaFold summary and mmcif file for given UniProt accessions.

    Args:
        uniprot_accs: A set of UniProt accessions.
        save_dir: The directory to save the fetched files.

    Returns:
        A list of AlphaFold entries.
    """
    what: set[DownloadableFormat] = {"summary", "cif"}
    return alphafold_fetch(uniprot_accs, save_dir, what)


@mcp.tool
async def fetch_emdb_volumes(emdb_ids: set[str], save_dir: Path) -> Mapping[str, Path]:
    """Fetch EMDB volumes for given EMDB IDs.

    Args:
        emdb_ids: A set of EMDB IDs.
        save_dir: The directory to save the fetched files.
    Returns:
        A mapping of EMDB ID to the path of the fetched volume file.
    """
    return await emdb_fetch(emdb_ids=emdb_ids, save_dir=save_dir)


@mcp.tool
def alphafold_confidence_filter(file: Path, query: ConfidenceFilterQuery, filtered_dir: Path) -> ConfidenceFilterResult:
    """Take a mmcif/PDB file and filter it based on confidence (plDDT) scores.

    If passes filter writes file to filtered_dir with residues above confidence threshold.
    """
    return filter_file_on_residues(file, query, filtered_dir)


mcp.tool(filter_file_on_secondary_structure)


@mcp.prompt
def candidate_structures(
    species: str = "Human",
    cellular_location: str = "nucleus",
    confidence: int = 90,
    min_residues: int = 100,
    max_residues: int = 200,
) -> str:
    """Prompt to find candidate structures.

    Args:
        species: The species to search for (default: "Human").
        cellular_location: The cellular location to search for (default: "nucleus").
        confidence: The confidence threshold for AlphaFold structures (default: 90).
        min_residues: Minimum number of high confidence residues (default: 100).
        max_residues: Maximum number of high confidence residues (default: 200).

    Returns:
        A prompt string to find candidate structures.
    """
    return dedent(f"""\
        Given the species '{species}' and cellular location '{cellular_location}' find the candidate structures.
        Download structures from 2 sources namely PDB and Alphafold.
        For alphafold I only want to use high confidence scores of over {confidence}.
        and only keep structures with number of high confidence residues between {min_residues} and {max_residues}.

        1. Search uniprot for proteins related to {species} and {cellular_location}.
            1. For the species find the NCBI taxonomy id.
            2. For cellular location find the associated GO term.
            3. Find uniprot accessions based on NCBI taxonomy id and cellular location GO term.
        2. For PDB
            1. Search for structures related to the identified proteins.
            2. Download each PDB entry from PDBe
            3. Extract chain for the protein of interest.
        3. For Alphafold
            1. Search for AlphaFold entries related to the identified proteins.
            2. Download each AlphaFold entry.
            3. Filter the structures based on {confidence} as confidence
               and nr residues between {min_residues} and {max_residues}.
    """)



"""Module for searching taxon information from UniProt."""

import csv
import gzip
import logging
from dataclasses import dataclass
from typing import Literal, get_args

from aiohttp.client import ClientResponse
from aiohttp_retry import RetryClient
from cattrs.gen import make_dict_structure_fn, override
from yarl import URL

from protein_quest.converter import converter
from protein_quest.go import TextIOWrapper
from protein_quest.utils import friendly_session

logger = logging.getLogger(__name__)


@dataclass(frozen=True, slots=True)
class Taxon:
    """Dataclass representing a taxon.

    Arguments:
        taxon_id: The unique identifier for the taxon.
        scientific_name: The scientific name of the taxon.
        rank: The taxonomic rank of the taxon (e.g., species, genus).
        common_name: The common name of the taxon (if available).
        other_names: A set of other names for the taxon (if available).
    """

    taxon_id: str
    scientific_name: str
    rank: str
    common_name: str | None = None
    other_names: set[str] | None = None


@dataclass(frozen=True, slots=True)
class SearchTaxonResponse:
    results: list[Taxon]


converter.register_structure_hook(
    Taxon,
    make_dict_structure_fn(
        Taxon,
        converter,
        taxon_id=override(rename="taxonId"),
        scientific_name=override(rename="scientificName"),
        common_name=override(rename="commonName"),
        other_names=override(rename="otherNames"),
    ),
)

SearchField = Literal["tax_id", "scientific", "common", "parent"]
"""Type of search field"""
search_fields: set[SearchField | None] = set(get_args(SearchField)) | {None}
"""Set of valid search fields"""


def _get_next_page(response: ClientResponse) -> URL | str | None:
    next_page = response.links.getone("next", None)
    if next_page is None:
        return None
    return next_page.getone("url", None)


async def _fetch_page(url: URL | str, session: RetryClient) -> tuple[list[Taxon], URL | str | None]:
    async with session.get(url) as response:
        response.raise_for_status()
        gzipped_raw_data = await response.read()
    next_page = _get_next_page(response)
    raw_data = gzip.decompress(gzipped_raw_data)
    taxons = converter.loads(raw_data, SearchTaxonResponse).results
    return taxons, next_page


async def search_taxon(query: str, field: SearchField | None = None, limit: int = 100) -> list[Taxon]:
    """Search for taxon information in UniProt.

    Uses <https://www.uniprot.org/taxonomy?query=*>.

    Args:
        query: Search query for the taxon.
        field: Field to search in.
            If None, searches in all fields.
            If "tax_id" then searches by taxon ID.
            If "parent" then given a parent taxon ID returns all its children.
            For example, if the parent taxon ID is 9606 (Human), it will return Neanderthal and Denisovan.
        limit: Maximum number of results to return.

    Returns:
        List of Taxon objects matching the search query.

    Raises:
        ValueError: If the search field is invalid.
    """
    # https://rest.uniprot.org/taxonomy/search?compressed=true&format=json&query=%28Mouse%29&size=500
    page_limit = 100
    if field not in search_fields:
        msg = f"Invalid search field: {field}. Must be one of {search_fields}."
        raise ValueError(msg)
    if field is not None:
        # ((common:"house+mouse"))
        query = f'(({field}:"{query}"))'
    params = {"query": query, "limit": str(page_limit), "compressed": "true", "format": "json"}
    url = URL("https://rest.uniprot.org/taxonomy/search").with_query(params)
    logger.debug("Fetching uniprot taxonomy from %s with params %s", url, params)
    async with friendly_session() as session:
        # Fetch first page
        taxons, next_page = await _fetch_page(url, session)
        if len(taxons) >= limit:
            return taxons[:limit]
        if next_page is None:
            return taxons

        # Fetch next pages
        while next_page:
            logger.debug("Fetching next page of uniprot taxonomy from %s", next_page)
            next_data, next_page = await _fetch_page(next_page, session)
            taxons.extend(next_data)
            if len(taxons) >= limit:
                return taxons[:limit]
            if next_page is None:
                return taxons
    return taxons


def _write_taxonomy_csv(taxons: list[Taxon], output_csv: TextIOWrapper) -> None:
    """Write taxon information to a CSV file.

    Args:
        taxons: List of Taxon objects to write to the CSV file.
        output_csv: File object for the output CSV file.
    """
    writer = csv.writer(output_csv)
    writer.writerow(["taxon_id", "scientific_name", "common_name", "rank", "other_names"])
    for taxon in taxons:
        writer.writerow(
            [
                taxon.taxon_id,
                taxon.scientific_name,
                taxon.common_name,
                taxon.rank,
                ";".join(taxon.other_names) if taxon.other_names else "",
            ]
        )



import logging
from pathlib import Path

import gemmi
import pytest

from protein_quest.pdbe.io import (
    ChainNotFoundError,
    glob_structure_files,
    locate_structure_file,
    nr_residues_in_chain,
    write_single_chain_pdb_file,
    write_structure,
)


@pytest.fixture
def cif_path() -> Path:
    return Path(__file__).parent / "fixtures" / "2y29.cif"


def test_write_single_chain_pdb_file_happypath(cif_path: Path, tmp_path: Path):
    output_file = write_single_chain_pdb_file(
        input_file=cif_path,
        chain2keep="A",
        output_dir=tmp_path,
        out_chain="Z",
    )

    assert output_file is not None
    assert output_file.name == "2y29_A2Z.cif"
    assert output_file.exists()
    structure = gemmi.read_structure(str(output_file))
    assert len(structure) == 1  # One model
    model = structure[0]
    assert len(model) == 1  # One chain
    chain = model[0]
    assert chain.name == "Z"
    assert len(chain) == 6  # 6 residues in chain Z


def test_write_single_chain_pdb_file_with_secondary_structure(tmp_path: Path):
    # See ../test_ss:sample_cif fixture how input_file was made
    input_file = Path(__file__).parent.parent / "fixtures" / "3JRS_B2A.cif.gz"
    output_file = write_single_chain_pdb_file(
        input_file=input_file,
        chain2keep="A",
        output_dir=tmp_path,
    )
    structure = gemmi.read_structure(str(output_file))
    assert len(structure.helices) == 4
    assert len(structure.sheets) == 1


def test_write_single_chain_pdb_file_already_exists(cif_path: Path, tmp_path: Path, caplog: pytest.LogCaptureFixture):
    fake_output_file = tmp_path / "2y29_A2Z.cif"
    fake_output_file.write_text("fake content")
    caplog.set_level(logging.INFO)

    output_file = write_single_chain_pdb_file(
        input_file=cif_path,
        chain2keep="A",
        output_dir=tmp_path,
        out_chain="Z",
    )

    assert output_file == fake_output_file
    assert "Skipping" in caplog.text


def test_write_single_chain_pdb_file_unknown_chain(cif_path: Path, tmp_path: Path):
    with pytest.raises(ChainNotFoundError):
        write_single_chain_pdb_file(
            cif_path,
            chain2keep="B",
            output_dir=tmp_path,
            out_chain="Z",
        )


def test_write_single_chain_pdb_file_unknown_format(tmp_path: Path):
    with pytest.raises(RuntimeError, match="Unknown format"):
        write_single_chain_pdb_file(
            tmp_path / "nonexistent_file.xyz",
            chain2keep="B",
            output_dir=tmp_path,
            out_chain="Z",
        )


def test_nr_residues_in_chain(cif_path: Path):
    residue_count = nr_residues_in_chain(cif_path, chain="A")

    assert residue_count == 8


def test_nr_residues_in_chain_wrongchain(cif_path: Path, caplog):
    residue_count = nr_residues_in_chain(cif_path, chain="Z")

    assert residue_count == 0
    assert "Chain Z not found in" in caplog.text


@pytest.mark.parametrize("extension", [".pdb", ".pdb.gz", ".cif", ".cif.gz"])
def test_write_structure(cif_path: Path, tmp_path: Path, extension: str):
    structure = gemmi.read_structure(str(cif_path))
    output_file = tmp_path / f"bla{extension}"

    write_structure(structure, output_file)

    found_files = list(glob_structure_files(tmp_path))
    assert len(found_files) == 1
    assert found_files[0].name == output_file.name


@pytest.mark.parametrize(
    "pdb_id, file_name",
    [
        # extensions
        ("2y29", "2y29.cif"),
        ("2y29", "2y29.cif.gz"),
        ("2y29", "2y29.pdb"),
        ("2y29", "2y29.pdb.gz"),
        ("2y29", "pdb2y29.ent"),
        ("2y29", "pdb2y29.ent.gz"),
        # cases
        ("1KVm", "1KVm.cif"),
        ("1KVm", "1kvm.cif"),
        ("1KVm", "1KVM.cif"),
    ],
)
def test_locate_structure_file(tmp_path: Path, pdb_id: str, file_name: str):
    test_input_file = tmp_path / file_name
    test_input_file.write_text("fake content")
    result = locate_structure_file(tmp_path, pdb_id)

    assert result == test_input_file


def test_locate_structure_file_notfound(tmp_path: Path):
    with pytest.raises(FileNotFoundError, match="No structure file found for nonexistent_id in"):
        locate_structure_file(tmp_path, "nonexistent_id")



import pytest
from yarl import URL

from protein_quest.converter import Percentage, PositiveInt, Ratio, converter

items = [
    ("https://example.com", URL, URL("https://example.com")),
    (42.0, Percentage, 42.0),
    (0.42, Ratio, 0.42),
    (42, PositiveInt, 42),
]


@pytest.mark.parametrize("raw,typ,expected", items)
def test_happy_structure(raw, typ, expected):
    res = converter.structure(raw, typ)
    assert res == expected


@pytest.mark.parametrize("expected,typ,raw", items)
def test_happy_unstructure(expected, typ, raw):
    res = converter.unstructure(raw)
    assert res == expected
